/*
 * denoise-dp.sdsl.cpp: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 * Copyright (C) 2009-2013 University of California Los Angeles
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: denoise-dp.sdsl.cpp
 * @author Pascal Getreuer <original author, UCLA Applied Math>
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 * @author: Justin Holewinski <justin.holewinski@gmail.com>
 */
/*
 * Implementation of Rician 3D pipeline stage using embedded DSL.
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/time.h>
#include <time.h>     

// Rician denoise SDSL version
int sdsl_rician3d(double *U, double *F, double sigma, double lambda, int dim0, int dim1, int dim2) {
  // G is a local array
  double *G = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  memset(G, 0, sizeof(double)*dim0*dim1*dim2);

  // Copy F to U
  memcpy(U, F, dim0*dim1*dim2*sizeof(double));
    
  // Run kernel
  #pragma sdsl begin gpu(block:8,8,8 tile:1,3,1 time:1) // gpu(...) is optional
  int dim0;
  int dim1;
  int dim2;

  double sigma;
  double lambda;
  
  double DT = 5.0;
  double EPSILON = 1.0E-20;
  
  double TOL = 0.00001;
  
  double sigma2 = sigma*sigma;
  double gamma = lambda/sigma2;
  
  grid g[dim2][dim1][dim0];
  
  double griddata U on g at 0,1;
  double griddata G on g at 0;
  double griddata F on g at 0;
  
  pointfunction approx_g(u,g) {
    [0]g[0][0][0] = 1.0 / sqrt(EPSILON +
                           ([0]u[0][0][0] - [0]u[ 0][ 1][ 0]) * ([0]u[0][0][0] - [0]u[ 0][ 1][ 0]) +
                           ([0]u[0][0][0] - [0]u[ 0][-1][ 0]) * ([0]u[0][0][0] - [0]u[ 0][-1][ 0]) +
                           ([0]u[0][0][0] - [0]u[ 0][ 0][ 1]) * ([0]u[0][0][0] - [0]u[ 0][ 0][ 1]) +
                           ([0]u[0][0][0] - [0]u[ 0][ 0][-1]) * ([0]u[0][0][0] - [0]u[ 0][ 0][-1]) +
                           ([0]u[0][0][0] - [0]u[ 1][ 0][ 0]) * ([0]u[0][0][0] - [0]u[ 1][ 0][ 0]) +
                           ([0]u[0][0][0] - [0]u[-1][ 0][ 0]) * ([0]u[0][0][0] - [0]u[-1][ 0][ 0]) );
  }
  
  pointfunction update_u(u,g,f) {
    double r = [0]u[0][0][0]*[0]f[0][0][0]/sigma2;
    r =( r*(2.38944 + r*(0.950037 + r)) )
           / ( 4.65314 + r*(2.57541 + r*(1.48937 + r)) );
    [1]u[0][0][0] =   ([0]u[0][0][0] +
                       DT*([0]u[ 0][ 1][ 0]*[0]g[ 0][ 1][ 0] +
                           [0]u[ 0][-1][ 0]*[0]g[ 0][-1][ 0] +
                           [0]u[ 0][ 0][ 1]*[0]g[ 0][ 0][ 1] +
                           [0]u[ 0][ 0][-1]*[0]g[ 0][ 0][-1] +
                           [0]u[ 1][ 0][ 0]*[0]g[ 1][ 0][ 0] +
                           [0]u[-1][ 0][ 0]*[0]g[-1][ 0][ 0] +
                           gamma*[0]f[0][0][0]*r))
                     /
                      (1.0 + DT*[0]g[ 0][ 1][ 0] +
                                [0]g[ 0][-1][ 0] +
                                [0]g[ 0][ 0][ 1] +
                                [0]g[ 0][ 0][-1] +
                                [0]g[ 1][ 0][ 0] +
                                [0]g[-1][ 0][ 0] +
                                gamma);
  }
  
  iterate 50 {
    stencil gs {
      [1:dim2-2][1:dim1-2][1:dim0-2] : approx_g(U,G); 
    }
    stencil us {
      [1:dim2-2][1:dim1-2][1:dim0-2] : update_u(U,G,F); 
    }
    
    reduction max_diff max {
      [1:dim2-2][1:dim1-2][1:dim0-2] : fabs([1]U[0][0][0] - [0]U[0][0][0]);
    }
    
  } check (max_diff < TOL) every 10 iterations
  #pragma sdsl end
  
  // Clean-up
  free(G);
  return sdsl_return_0;
}

#ifdef CHECK_REF
// Rician denoise reference version
int rician3d(double *U, double *F, double sigma, double lambda, int dim0, int dim1, int dim2) {
  // Maximum iterations
  const int max_iter = 50;
  
  // U0, G are local arrays
  double *U0 = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  double *G = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  memset(G, 0, sizeof(double)*dim0*dim1*dim2);

  
  // Copy F to U
  memcpy(U, F, dim0*dim1*dim2*sizeof(double));
  
  // Constants
  double DT = 5.0;
  double EPSILON = 1.0E-20;
  double TOL = 0.00001;
  double sigma2 = sigma*sigma;
  double gamma = lambda/sigma2;
  
  // Convergence
  bool converged = true;
  
  // Macro for 3d indexing
  #define idx(z,y,x) ((x) + (y)*dim0 + (z)*dim0*dim1)
  /* Gradient descent loop */
  for (int t = 0; t < max_iter; ++t) {
    
    /* Copy U to U0 */
    for (int i = 0; i < dim2; ++i) {
      for (int j = 0; j < dim1; ++j) {
        for (int k = 0; k < dim0; ++k) {
          U0[idx(i,j,k)] = U[idx(i,j,k)];
        }
      }
    }
    
    /* Approximate G = 1/|grad U| */
    for (int i = 1; i < dim2 - 1; ++i) {
      for (int j = 1; j < dim1 - 1; ++j) {
        for (int k = 1; k < dim0 - 1; ++k) {
          G[idx(i,j,k)] = 1.0/sqrt(EPSILON + 
          (U0[idx(i,j,k)] - U0[idx(i,j+1,k)])*(U0[idx(i,j,k)] - U0[idx(i,j+1,k)]) + 
          (U0[idx(i,j,k)] - U0[idx(i,j-1,k)])*(U0[idx(i,j,k)] - U0[idx(i,j-1,k)]) + 
          (U0[idx(i,j,k)] - U0[idx(i,j,k+1)])*(U0[idx(i,j,k)] - U0[idx(i,j,k+1)]) + 
          (U0[idx(i,j,k)] - U0[idx(i,j,k-1)])*(U0[idx(i,j,k)] - U0[idx(i,j,k-1)]) + 
          (U0[idx(i,j,k)] - U0[idx(i+1,j,k)])*(U0[idx(i,j,k)] - U0[idx(i+1,j,k)]) + 
          (U0[idx(i,j,k)] - U0[idx(i-1,j,k)])*(U0[idx(i,j,k)] - U0[idx(i-1,j,k)]));
        }
      }
    }
    
    /* Update U by a semi-implict step */
    converged = true;
    for (int i = 1; i < dim2 - 1; ++i) {
      for (int j = 1; j < dim1 - 1; ++j) {
        for (int k = 1; k < dim0 - 1; ++k) {
          /* Evaluate r = I1(U*F/sigma^2) / I0(U*F/sigma^2) 
             with a cubic rational approximation. */
          double r = U[idx(i,j,k)]*F[idx(i,j,k)]/sigma2;
          r = ( r*(2.38944 + r*(0.950037 + r)) )
          / ( 4.65314 + r*(2.57541 + r*(1.48937 + r)) );
          /* Update U */
          U[idx(i,j,k)] = (U0[idx(i,j,k)] +
                             DT*(U0[idx(i,j+1,k)]*G[idx(i,j+1,k)] +
                                 U0[idx(i,j-1,k)]*G[idx(i,j-1,k)] +
                                 U0[idx(i,j,k+1)]*G[idx(i,j,k+1)] +
                                 U0[idx(i,j,k-1)]*G[idx(i,j,k-1)] +
                                 U0[idx(i+1,j,k)]*G[idx(i+1,j,k)] +
                                 U0[idx(i-1,j,k)]*G[idx(i-1,j,k)] +
                                 gamma*F[idx(i,j,k)]*r))
                           /
                            (1.0 + DT*G[idx(i,j+1,k)] +
                                      G[idx(i,j-1,k)] +
                                      G[idx(i,j,k+1)] +
                                      G[idx(i,j,k-1)] +
                                      G[idx(i+1,j,k)] +
                                      G[idx(i-1,j,k)] +
                                      gamma);
          if (fabs(U0[idx(i,j,k)] - U[idx(i,j,k)]) > TOL) {
            converged = false;
          }
        }
      }
    }
    if (converged) {
      break;
    }
  }
  #undef idx
  free(U0);
  free(G);
  return 0;
}

// Correctness check
void check(double* ref, double* sdsl, int dim0, int dim1, int dim2) {
  double err_L2_norm = 0.0;
  double ref_L2_norm = 0.0;
  #ifdef DEBUG
  int num_diff = 0;
  #endif
                
  // Macro for 3d indexing
  #define idx(z,y,x) ((x) + (y)*dim0 + (z)*dim0*dim1) 
  for (int i = 1; i < dim2 - 1; ++i) {
    for (int j = 1; j < dim1 - 1; ++j) {
      for (int k = 1; k < dim0 - 1; ++k) {
        double diff = sdsl[idx(i,j,k)] - ref[idx(i,j,k)];
        err_L2_norm += diff*diff;
        ref_L2_norm += ref[idx(i,j,k)]*ref[idx(i,j,k)]; 
        #ifdef DEBUG
        if (diff != 0.0) {
          num_diff++;
          //printf("Diff at U[%d][%d][%d] = U[%d] = %.6E\n",i,j,k,idx(i,j,k),diff);
          //printf("UREF  = %.6E\n", ref[idx(i,j,k)]);
          //printf("USDSL = %.6E\n", sdsl[idx(i,j,k)]);
        }
        #endif
      }
    }
  }
  err_L2_norm = sqrt(err_L2_norm);
  ref_L2_norm = sqrt(ref_L2_norm);
  
  #ifdef DEBUG
  printf("Diff at %d cells\n",num_diff);
  #endif
  printf("L2 Norm (err): %.6E\n",err_L2_norm);
  printf("L2 Norm (ref): %.6E\n",ref_L2_norm);
  if (err_L2_norm/ref_L2_norm > 1e-2) {
    printf("FAIL\n");
  } else {
    printf("PASS\n");
  }
  #undef idx
}
#endif

#ifdef STANDALONE_DRIVER

// Current timestamp (s)
double timestamp() {
  struct timeval tim;
  gettimeofday(&tim,NULL);
  return tim.tv_sec + (tim.tv_usec/1000000.0);
}

int main() {
  // Problem size
  const int dim0 = 128;
  const int dim1 = 128;
  const int dim2 = 128;
  
  // Timing
  double start, finish;
  double sdsl_GF;
  #ifdef CHECK_REF
  double ref_GF;
  #endif
  
  // Data arrays
  double *U = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  double *F = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  #ifdef CHECK_REF
  double *Uref = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  double *Fref = (double*)malloc(dim0*dim1*dim2*sizeof(double));
  #endif

  // Populate arrays with random values
  srand(45);
  for (int i = 0; i < dim0*dim1*dim2; ++i) {
    F[i] = (double)rand() / (double)(RAND_MAX);
    #ifdef CHECK_REF
    Fref[i] = F[i];
    #endif
  }
  
  #ifdef CHECK_REF
  // Reference denoise call
  start = timestamp();
  int ref_ret = rician3d(Uref, Fref, 0.05, 0.065, dim0, dim1, dim2);
  finish = timestamp();
  printf(" REF: %.6f sec\n",finish-start);
  #endif
  
  // SDSL denoise call
  start = timestamp();
  int sdsl_ret = sdsl_rician3d(U, F, 0.05, 0.065, dim0, dim1, dim2);
  finish = timestamp();
  printf("SDSL: %.6f sec\n",finish-start);

  #ifdef CHECK_REF
  // Check results
  check(Uref,U,dim0,dim1,dim2);
  #endif
  
  // Cleanup
  free(U);
  free(F);
  #ifdef CHECK_REF
  free(Uref);
  free(Fref);
  #endif
  return sdsl_ret;
}
#endif
