/*
 * pipeline-sp.sdsl.cpp: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 * Copyright (C) 2001-2013 University of California Los Angeles
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: pipeline-sp.sdsl.cpp
 * @author: Tony Chan <Member, IEEE>
 * @author: Igor Yanovsky <UCLA Applied Math>
 * @author: Luminita Vese <UCLA Applied Math>
 * @author: Justin Holewinski <justin.holewinski@gmail.com>
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
#include <stdio.h>
#include <sys/time.h>
#include <cstring>
#include <cstdlib>
#include <cmath>



//===----------------------------------------------------------------------===//
// Initialization and utilities
//===----------------------------------------------------------------------===//

namespace {
	const int ImageN = 256;
	const int ImageM = 256;
	const int ImageP = 256;
	
	float *img;
	
	// Taken (and modified) from cpu_kernel.c
	void init_phi_sdsl(float* phi/*,float* curvature_motion_part*/)
	{
		int m,n,p;
		m = n=p=256;

		//
		//  This routine inputs the initial condition for phi.
		//
		float epsilon = 10e-5;
		long  i; 
		long  j;
		long  k;
		float xcent   = (m-1) / 2.0;
		float ycent   = (n-1) / 2.0;
		float zcent   = (p-1) / 2.0;
		float r       = (fmin(fmin((float)m,(float)n), (float)p)) / 3.0f;
		float xx; float yy; float zz;

		for(i=0; i < m; i++)
		{
			for(j=0; j < n; j++)
			{
				for(k=0; k < p; k++)
				{
					xx = (float)(i);
					yy = (float)(j);
					zz = (float)(k);
					phi[(i*n+j)*p+k] = sqrtf( (xx-xcent)*(xx-xcent) + (yy-ycent)*(yy-ycent) + (zz-zcent)*(zz-zcent)) - r;
				}
			}
		}
	}
}

#define SQR(x)       ((x)*(x))
#define REF3D(i,j,k) ((i)*dim0*dim1+(j)*dim0+(k))



//===----------------------------------------------------------------------===//
// Segmentation
//===----------------------------------------------------------------------===//

extern "C"
int seg_sdsl(int ImageID, float *u0) {
  printf("Starting SDSL segmentation...\n");

	const int IterPerKernel = 10;
	const int MaxIter       = 50;

	const int dim0 = ImageM;
	const int dim1 = ImageN;
	const int dim2 = ImageP;

	int ret = 0;
	// Initialization taken from cpu_kernel.c
	float* phi                   = (float*)calloc(dim0*dim1*dim2,sizeof(float));
	float* curvature_motion_part = (float*)calloc(dim0*dim1*dim2,sizeof(float));

	init_phi_sdsl(phi);

	float lambda1 = 1;
	float lambda2 = 1;
	float mu      = 0.18*255*255;
	float dt      = 0.15/mu;


	float num1, num2, den1, den2, c1, c2;

	num1 = 0.0;
	num2 = 0.0;
	den1 = 0.0;
	den2 = 0.0;

	for (int i = 0; i < MaxIter; i += IterPerKernel) {
		for(int i = 0; i < dim2; i++) {
			for(int j = 0; j < dim1; j++) {
				for(int k = 0; k < dim0; k++) {
					num1 = (phi[REF3D(i,j,k)] < 0) ? num1 + u0[REF3D(i,j,k)] : (num1);
					den1 = (phi[REF3D(i,j,k)] < 0) ? den1 + 1 : (den1);
					num2 = (phi[REF3D(i,j,k)] > 0) ? num2 + u0[REF3D(i,j,k)] : (num2);
					den2 = (phi[REF3D(i,j,k)] > 0) ? den2 + 1 : (den2);
				}
			}
		}

		c1 = num1 / den1;
		c2 = num2 / den2;

#pragma sdsl begin gpu(block:8,8,8 tile:1,1,1 time:1)

		int dim0;
		int dim1;
		int dim2;

		float c1;
		float c2;

		float dt;
		float mu;
		float lambda1;
		float lambda2;

		float EPSILON = 5e-5;

		grid g[dim2][dim1][dim0];

		float griddata u0  on g at 0,1;
		float griddata phi on g at 0,1;

		pointfunction calc_phi(p,u) {
			float Dx_p = ([0]p[1][0][0] - [0]p[ 0][ 0][ 0]);
			float Dx_m = ([0]p[0][0][0] - [0]p[-1][ 0][ 0]);
			float Dy_p = ([0]p[0][1][0] - [0]p[ 0][ 0][ 0]);
			float Dy_m = ([0]p[0][0][0] - [0]p[ 0][-1][ 0]);
			float Dz_p = ([0]p[0][0][1] - [0]p[ 0][ 0][ 0]);
			float Dz_m = ([0]p[0][0][0] - [0]p[ 0][ 0][-1]);

			float Dx_0 = ([0]p[1][0][0] - [0]p[-1][ 0][ 0])/2.0;
			float Dy_0 = ([0]p[0][1][0] - [0]p[ 0][-1][ 0])/2.0;
			float Dz_0 = ([0]p[0][0][1] - [0]p[ 0][ 0][-1])/2.0;

			float Dxx = (Dx_p - Dx_m);
			float Dyy = (Dy_p - Dy_m);
			float Dzz = (Dz_p - Dz_m);

			float Dxy = ([0]p[1][1][0] - [0]p[1][-1][0] - [0]p[-1][1][0] + [0]p[-1][-1][0] ) / 4.0;
			float Dxz = ([0]p[1][0][1] - [0]p[1][0][-1] - [0]p[-1][0][1] + [0]p[-1][0][-1] ) / 4.0;
			float Dyz = ([0]p[0][1][1] - [0]p[0][1][-1] - [0]p[0][-1][1] + [0]p[0][-1][-1] ) / 4.0;


			float K = (  Dx_0*Dx_0*Dyy - 2*Dx_0*Dy_0*Dxy + Dy_0*Dy_0*Dxx
					+ Dx_0*Dx_0*Dzz - 2*Dx_0*Dz_0*Dxz + Dz_0*Dz_0*Dxx
					+ Dy_0*Dy_0*Dzz - 2*Dy_0*Dz_0*Dyz + Dz_0*Dz_0*Dyy )
				/ ( powf(Dx_0*Dx_0 + Dy_0*Dy_0 + Dz_0*Dz_0, 1.5) + EPSILON);

			float Grad = sqrtf(Dx_0*Dx_0 + Dy_0*Dy_0 + Dz_0*Dz_0);

			[1]p[0][0][0]	= [0]p[0][0][0] + dt * Grad * (mu * K + lambda1*SQR([0]u0[0][0][0]-c1) - lambda2*SQR([0]u0[0][0][0]-c2));
		}


		iterate 10 { // Hardcoded IterPerKernel
			stencil segi {
				[1:dim2-2][1:dim1-2][1:dim0-2] : calc_phi(phi, u0);
				[0:dim2-1][0:dim1-1][0     ] : [1]phi[0][0][0] = 0.0;
				[0:dim2-1][0:dim1-1][dim0-1  ] : [1]phi[0][0][0] = 0.0;

				[0:dim2-1][0     ][0:dim0-1] : [1]phi[0][0][0] = 0.0;
				[0:dim2-1][dim1-1  ][0:dim0-1] : [1]phi[0][0][0] = 0.0;

				[0     ][0:dim1-1][0:dim0-1] : [1]phi[0][0][0] = 0.0;
				[dim2-1  ][0:dim1-1][0:dim0-1] : [1]phi[0][0][0] = 0.0;
			}
		} 
#pragma sdsl end
    ret = sdsl_return_0;
	}

	free(phi);
	free(curvature_motion_part);
  printf("Finished SDSL segmentation.\n");
	return ret;
}


//===----------------------------------------------------------------------===//
// Rician Denoise
//===----------------------------------------------------------------------===//
extern "C"
float* denoise_sdsl(int ImageID, float *F) {
  printf("Starting SDSL denoise...\n");
	const int dim0 = ImageM;
	const int dim1 = ImageN;
	const int dim2 = ImageP;

	const float sigma     = 0.05f;
	const float lambda    = 0.065f;
	const float Tolerance = 2e-3f;

	const int MaxIter     = 50;
	const int IterPerCall = 1;

	// Copy F into work array U
	// Use C malloc() since we have to pass the pointer back to the caller
	float *U = (float*)malloc(sizeof(float)*dim0*dim1*dim2);
	memcpy(U, F, sizeof(float)*dim0*dim1*dim2);


		// Run kernel
#pragma sdsl begin gpu(block:8,8,8 tile:1,3,1 time:1)
		int dim0;
		int dim1;
		int dim2;

		float sigma;
		float lambda;

		float Tolerance;

		float DT = 5.0;
		float EPSILON = 1.0E-20;

		float sigma2 = sigma*sigma;
		float gamma  = lambda/sigma2;

		grid g[dim2][dim1][dim0];

		float griddata U on g at 0,1;
		float griddata F on g at 0;

		pointfunction update_u(u,f) {
			float tmp = [0]u[0][0][0]*[0]f[0][0][0]/sigma2;
			float r =   ( tmp*(2.38944 + tmp*(0.950037 + tmp)) )
				/ ( 4.65314 + tmp*(2.57541 + tmp*(1.48937 + tmp)) );


			float g_m1_0_0 = rsqrtf(EPSILON +
					([0]u[-1][0][0] - [0]u[ -1][ 1][ 0])*([0]u[-1][0][0] - [0]u[ -1][ 1][ 0]) +
					([0]u[-1][0][0] - [0]u[ -1][-1][ 0])*([0]u[-1][0][0] - [0]u[ -1][-1][ 0]) +
					([0]u[-1][0][0] - [0]u[ -1][ 0][ 1])*([0]u[-1][0][0] - [0]u[ -1][ 0][ 1]) +
					([0]u[-1][0][0] - [0]u[ -1][ 0][-1])*([0]u[-1][0][0] - [0]u[ -1][ 0][-1]) +
					([0]u[-1][0][0] - [0]u[  0][ 0][ 0])*([0]u[-1][0][0] - [0]u[  0][ 0][ 0]) +
					([0]u[-1][0][0] - [0]u[ -2][ 0][ 0])*([0]u[-1][0][0] - [0]u[ -2][ 0][ 0]) );

			float g_p1_0_0 = rsqrtf(EPSILON +
					([0]u[1][0][0] - [0]u[ 1][ 1][ 0])*([0]u[1][0][0] - [0]u[ 1][ 1][ 0]) +
					([0]u[1][0][0] - [0]u[ 1][-1][ 0])*([0]u[1][0][0] - [0]u[ 1][-1][ 0]) +
					([0]u[1][0][0] - [0]u[ 1][ 0][ 1])*([0]u[1][0][0] - [0]u[ 1][ 0][ 1]) +
					([0]u[1][0][0] - [0]u[ 1][ 0][-1])*([0]u[1][0][0] - [0]u[ 1][ 0][-1]) +
					([0]u[1][0][0] - [0]u[ 2][ 0][ 0])*([0]u[1][0][0] - [0]u[ 2][ 0][ 0]) +
					([0]u[1][0][0] - [0]u[ 0][ 0][ 0])*([0]u[1][0][0] - [0]u[ 0][ 0][ 0]) );

			float g_0_m1_0 = rsqrtf(EPSILON +
					([0]u[0][-1][0] - [0]u[ 0][ 0][ 0])*([0]u[0][-1][0] - [0]u[ 0][ 0][ 0]) +
					([0]u[0][-1][0] - [0]u[ 0][-2][ 0])*([0]u[0][-1][0] - [0]u[ 0][-2][ 0]) +
					([0]u[0][-1][0] - [0]u[ 0][-1][ 1])*([0]u[0][-1][0] - [0]u[ 0][-1][ 1]) +
					([0]u[0][-1][0] - [0]u[ 0][-1][-1])*([0]u[0][-1][0] - [0]u[ 0][-1][-1]) +
					([0]u[0][-1][0] - [0]u[ 1][-1][ 0])*([0]u[0][-1][0] - [0]u[ 1][-1][ 0]) +
					([0]u[0][-1][0] - [0]u[-1][-1][ 0])*([0]u[0][-1][0] - [0]u[-1][-1][ 0]) );

			float g_0_p1_0 = rsqrtf(EPSILON +
					([0]u[0][1][0] - [0]u[ 0][ 2][ 0])*([0]u[0][1][0] - [0]u[ 0][ 2][ 0]) +
					([0]u[0][1][0] - [0]u[ 0][ 0][ 0])*([0]u[0][1][0] - [0]u[ 0][ 0][ 0]) +
					([0]u[0][1][0] - [0]u[ 0][ 1][ 1])*([0]u[0][1][0] - [0]u[ 0][ 1][ 1]) +
					([0]u[0][1][0] - [0]u[ 0][ 1][-1])*([0]u[0][1][0] - [0]u[ 0][ 1][-1]) +
					([0]u[0][1][0] - [0]u[ 1][ 1][ 0])*([0]u[0][1][0] - [0]u[ 1][ 1][ 0]) +
					([0]u[0][1][0] - [0]u[-1][ 1][ 0])*([0]u[0][1][0] - [0]u[-1][ 1][ 0]) );

			float g_0_0_m1 = rsqrtf(EPSILON +
					([0]u[0][0][-1] - [0]u[ 0][ 1][ -1])*([0]u[0][0][-1] - [0]u[ 0][ 1][-1]) +
					([0]u[0][0][-1] - [0]u[ 0][-1][ -1])*([0]u[0][0][-1] - [0]u[ 0][-1][-1]) +
					([0]u[0][0][-1] - [0]u[ 0][ 0][  0])*([0]u[0][0][-1] - [0]u[ 0][ 0][ 0]) +
					([0]u[0][0][-1] - [0]u[ 0][ 0][ -2])*([0]u[0][0][-1] - [0]u[ 0][ 0][-2]) +
					([0]u[0][0][-1] - [0]u[ 1][ 0][ -1])*([0]u[0][0][-1] - [0]u[ 1][ 0][-1]) +
					([0]u[0][0][-1] - [0]u[-1][ 0][ -1])*([0]u[0][0][-1] - [0]u[-1][ 0][-1]) );

			float g_0_0_p1 = rsqrtf(EPSILON +
					([0]u[0][0][1] - [0]u[ 0][ 1][ 1])*([0]u[0][0][1] - [0]u[ 0][ 1][ 1]) +
					([0]u[0][0][1] - [0]u[ 0][-1][ 1])*([0]u[0][0][1] - [0]u[ 0][-1][ 1]) +
					([0]u[0][0][1] - [0]u[ 0][ 0][ 2])*([0]u[0][0][1] - [0]u[ 0][ 0][ 2]) +
					([0]u[0][0][1] - [0]u[ 0][ 0][ 0])*([0]u[0][0][1] - [0]u[ 0][ 0][ 0]) +
					([0]u[0][0][1] - [0]u[ 1][ 0][ 1])*([0]u[0][0][1] - [0]u[ 1][ 0][ 1]) +
					([0]u[0][0][1] - [0]u[-1][ 0][ 1])*([0]u[0][0][1] - [0]u[-1][ 0][ 1]) );


			[1]u[0][0][0] = ([0]u[0][0][0] +
					DT*([0]u[ 0][ 1][ 0]*g_0_p1_0 +
						[0]u[ 0][-1][ 0]*g_0_m1_0 +
						[0]u[ 0][ 0][ 1]*g_0_0_p1 +
						[0]u[ 0][ 0][-1]*g_0_0_m1 +
						[0]u[ 1][ 0][ 0]*g_p1_0_0 +
						[0]u[-1][ 0][ 0]*g_m1_0_0 +
						gamma*[0]f[0][0][0]*r))
				/
				(1.0 + DT*(g_0_p1_0 +
									 g_0_m1_0 +
									 g_0_0_p1 +
									 g_0_0_m1 +
									 g_p1_0_0 +
									 g_m1_0_0 +
									 gamma));
		}

		iterate 50 { // hardcoded MaxIter
			stencil us {
				[2:dim2-3][2:dim1-3][2:dim0-3] : update_u(U, F); 
			}
      reduction max_diff max {
        [1:dim2-2][1:dim1-2][1:dim0-2] : fabs([1]U[0][0][0] - [0]U[0][0][0]);
      }
		} check (max_diff < Tolerance) every 1 iterations // Hardcoded IterPerCall
#pragma sdsl end

  printf("Finished SDSL denoise.\n");
	return U;
}


//===----------------------------------------------------------------------===//
// Registration
//===----------------------------------------------------------------------===//

// CPU U_cal
#define addrCal(a,b,c) ((a)*n+(b))*p+(c)
#define u1(i,j,k)      U1[addrCal(i,j,k)]
#define u2(i,j,k)      U2[addrCal(i,j,k)]
#define u3(i,j,k)      U3[addrCal(i,j,k)]
#define v1(i,j,k)      V1[addrCal(i,j,k)]
#define v2(i,j,k)      V2[addrCal(i,j,k)]
#define v3(i,j,k)      V3[addrCal(i,j,k)]
#define interpT(i,j,k) interpT[addrCal(i,j,k)]
#define S(i,j,k)       S[addrCal(i,j,k)]
#define T(i,j,k)       T[addrCal(i,j,k)]

namespace {
	void Ucal_CPU(float* V1,float* V2,float *V3,float* U1,float* U2,float* U3,float* S,float* T, float* interpT,int iter)
	{
		int i,j,k;
		float Rmax_L2=0.0;
		int m,n,p;
		m=n=p=256;
#pragma omp parallel for private(j,k) schedule(dynamic)
		for( i = 1; i < m-1; i++ )
		{
			for( j = 1; j < n-1; j++ )
			{
				for( k = 1; k < p-1; k++ )
				{
					float   du1_dx = (u1(i+1,j,k) - u1(i-1,j,k)) *0.5f;
					float   du2_dx = (u2(i+1,j,k) - u2(i-1,j,k)) *0.5f;
					float   du3_dx = (u3(i+1,j,k) - u3(i-1,j,k)) *0.5f;

					float   du1_dy = (u1(i,j+1,k) - u1(i,j-1,k)) *0.5f;
					float   du2_dy = (u2(i,j+1,k) - u2(i,j-1,k)) *0.5f;
					float   du3_dy = (u3(i,j+1,k) - u3(i,j-1,k)) *0.5f;

					float   du1_dz = (u1(i,j,k+1) - u1(i,j,k-1)) *0.5f;
					float   du2_dz = (u2(i,j,k+1) - u2(i,j,k-1)) *0.5f;
					float   du3_dz = (u3(i,j,k+1) - u3(i,j,k-1)) *0.5f;

					float r1 = v1(i,j,k) - v1(i,j,k)*du1_dx - v2(i,j,k)*du1_dy - v3(i,j,k)*du1_dz;
					float r2 = v2(i,j,k) - v1(i,j,k)*du2_dx - v2(i,j,k)*du2_dy - v3(i,j,k)*du2_dz;
					float r3 = v3(i,j,k) - v1(i,j,k)*du3_dx - v2(i,j,k)*du3_dy - v3(i,j,k)*du3_dz;

					v1(i,j,k)=r1;
					v2(i,j,k)=r2;
					v3(i,j,k)=r3;

					//float  R_L2 = sqrtf(r1*r1+r2*r2+r3*r3 );
					//Rmax_L2 = (Rmax_L2 > R_L2 ) ? Rmax_L2 : R_L2;
				}
			}
		}
		//float 	dt = 0.1/Rmax_L2;
		//float dt=0.0001625f*PostScale;
		//float dt=0.0001625f;
		float dt=1.0/(1<<12);
		//printf("dt = %f\n",dt);

#pragma omp parallel for private(j,k) schedule(dynamic)
		for( i = 1; i < m-1; i++ )
			for( j = 1; j < n-1; j++ )
				for( k = 1; k < p-1; k++ )
				{
					u1(i,j,k)+=dt*v1(i,j,k); 
					u2(i,j,k)+=dt*v2(i,j,k); 
					u3(i,j,k)+=dt*v3(i,j,k); 
				}

#pragma omp parallel for private(j,k) schedule(dynamic)
		for(i=1;i<m-1;i++)
			for(j=1;j<n-1;j++)
				for(k=1;k<p-1;k++)
				{
					float itmp_f=i-u1(i,j,k);
					int itmp=(int)itmp_f;
					float jtmp_f=j-u2(i,j,k);
					int jtmp=(int)jtmp_f;
					float ktmp_f=k-u3(i,j,k);
					int ktmp=(int)ktmp_f;

					interpT(i,j,k)= (1.0f+itmp-itmp_f)*(1.0f+jtmp-jtmp_f)*(1.0f+ktmp-ktmp_f)*T(itmp,jtmp,ktmp)+
						(1.0f+itmp-itmp_f)*(1.0f+jtmp-jtmp_f)*(ktmp_f-ktmp)*T(itmp,jtmp,ktmp+1)+
						(1.0f+itmp-itmp_f)*(jtmp_f-jtmp)*(1.0f+ktmp-ktmp_f)*T(itmp,jtmp+1,ktmp)+
						(1.0f+itmp-itmp_f)*(jtmp_f-jtmp)*(ktmp_f-ktmp)*T(itmp,jtmp+1,ktmp+1)+
						(itmp_f-itmp)*(1.0f+jtmp-jtmp_f)*(1.0f+ktmp-ktmp_f)*T(itmp+1,jtmp,ktmp)+
						(itmp_f-itmp)*(1.0f+jtmp-jtmp_f)*(ktmp_f-ktmp)*T(itmp+1,jtmp,ktmp+1)+
						(itmp_f-itmp)*(jtmp_f-jtmp)*(1.0f+ktmp-ktmp_f)*T(itmp+1,jtmp+1,ktmp)+
						(itmp_f-itmp)*(jtmp_f-jtmp)*(ktmp_f-ktmp)*T(itmp+1,jtmp+1,ktmp+1);

				}

		double SSD=0.0;
		for( i = 1; i < m-1; i++ )
			for( j = 1; j < n-1; j++ )		// unscaled SSD;
		for( k = 1; k < p-1; k++ )	// Calculated at beginning of each iteration
			SSD += 0.5*SQR(interpT(i,j,k)-S(i,j,k));	

		printf("Now in CPU registration iter %d,  current SSD:%lf\r\n", iter,SSD /(m-2.0f)/(n-2.0f)/(p-2.0f));

		//if(SSD<50) exit(0);

#pragma omp parallel for private(j,k) schedule(dynamic)
		for( i = 1; i < m-1; i++ )
		{
			for( j = 1; j < n-1; j++ )
			{
				for( k = 1; k < p-1; k++ )
				{
					float		dTdx = (interpT(i+1,j,k) - interpT(i-1,j,k)) *0.5f;
					float		dTdy = (interpT(i,j+1,k) - interpT(i,j-1,k)) *0.5f;
					float		dTdz = (interpT(i,j,k+1) - interpT(i,j,k-1)) *0.5f;

					v1(i,j,k) = (interpT(i,j,k) - S(i,j,k) ) * dTdx;
					v2(i,j,k) = (interpT(i,j,k) - S(i,j,k) ) * dTdy;
					v3(i,j,k) = (interpT(i,j,k) - S(i,j,k) ) * dTdz;
				}
			}
		}
	}
}

// Code taken from cpu_kernel.c, with V_cal replaced by SDSL
	extern "C"
float* reg_sdsl(int imageID,float* denoise_output)
{
  printf("Starting SDSL registration...\n");
	const int dim0 = ImageM;
	const int dim1 = ImageN;
	const int dim2 = ImageP;

	struct timeval  tim1,tim2;
	gettimeofday(&tim1, NULL);
	double          t1,t2;
	t1                = tim1.tv_sec+(tim1.tv_usec/1000000.0);
	printf("TIMESTAMP:reg sdsl for %d starts at %lf\r\n",imageID,t1);
	printf("SDSL kernel now working on image %d \r\n",imageID);
#if 1
	int             m,n,p;
	m                 = n=p=256;
	int             i,j,k;
	float          *T = (float*)calloc(m*n*p,sizeof(float));
	float          *S = (float*)calloc(m*n*p,sizeof(float));
	FILE*           S_ptr;
	FILE*           T_ptr;

	float           Smax = 8091.0, Tmax = 8180.0;
	S_ptr                = fopen("../data/S.img", "rb");
	unsigned short  num;
	unsigned short* buffer;
	buffer               = (unsigned short*)calloc(m*n*p,sizeof(unsigned short));
	fread(buffer,2,m*n*p,S_ptr);

	float t;
	for(k = 0; k < p; k++)
		for( j = 0; j < n; j++)
			for( i = 0; i < m; i++)
			{
				//fread(&num, 2, 1, S_ptr);
				t              = (float)buffer[(i*n+j)*p+k]/Smax*255.0f;
				S[(i*n+j)*p+k] = t;
			}
	fclose(S_ptr);
	free(buffer);
	for(k = 0; k < p; k++)
		for( j = 0; j < n; j++)
			for( i = 0; i < m; i++)
			{
				//fread(&num, 2, 1, T_ptr);
				//t            = (float)num/Tmax*255.0f;
				T[(i*n+j)*p+k] = denoise_output[(i*n+j)*p+k];
			}

	float *interpT = (float*)calloc(m*n*p,sizeof(float));

	float *U1      = (float*)calloc(m*n*p,sizeof(float));
	float *U2      = (float*)calloc(m*n*p,sizeof(float));
	float *U3      = (float*)calloc(m*n*p,sizeof(float));
	float *V1      = (float*)calloc(m*n*p,sizeof(float));
	float *V2      = (float*)calloc(m*n*p,sizeof(float));
	float *V3      = (float*)calloc(m*n*p,sizeof(float));
	int    iterNum = 100;

	long  GAUSSIAN_STEP = 3;
	float std           = 4.5;

	float lambda       = (std*std)/(2*GAUSSIAN_STEP);
	float nu           = (1 + 2*lambda - sqrt(1+4*lambda))/(2*lambda);
	float BoundryScale = 1/(1-nu);
	float PostScale    = pow(nu/lambda,3*GAUSSIAN_STEP);


	for(i=0;i<iterNum;i++)
	{
		// V_cal(V1,V2,V3)
#pragma sdsl begin gpu(block:8,8,8 tile:1,1,1 time:1)

		int dim0;
		int dim1;
		int dim2;

		float std;

		float lambda;
		float nu;
		float BoundryScale;
		float PostScale;

		grid g[dim2][dim1][dim0];

		float griddata V1 on g at 0,1;
		float griddata V2 on g at 0,1;
		float griddata V3 on g at 0,1;


		pointfunction filter_bnd(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [0]r1[0][0][0]*BoundryScale;
			[1]w2[0][0][0] = [0]r2[0][0][0]*BoundryScale;
			[1]w3[0][0][0] = [0]r3[0][0][0]*BoundryScale;
		}

		pointfunction filter_p(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [1]r1[0][0][0]*PostScale;
			[1]w2[0][0][0] = [1]r2[0][0][0]*PostScale;
			[1]w3[0][0][0] = [1]r3[0][0][0]*PostScale;
		}

		pointfunction filter_7(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [0]r1[ 0][ 0][ 0] + 
				([0]r1[ 0][ 0][-1] + 
				 [0]r1[ 0][ 0][ 1] + 
				 [0]r1[ 0][-1][ 0] + 
				 [0]r1[-1][ 1][ 0] + 
				 [0]r1[ 1][ 0][ 0] + 
				 [0]r1[0][0][ 0])*nu;
			[1]w2[0][0][0] = [0]r2[ 0][ 0][ 0] + 
				([0]r2[ 0][ 0][-1] + 
				 [0]r2[ 0][ 0][ 1] + 
				 [0]r2[ 0][-1][ 0] + 
				 [0]r2[-1][ 1][ 0] + 
				 [0]r2[ 1][ 0][ 0] + 
				 [0]r2[0][0][ 0])*nu;
			[1]w2[0][0][0] = [0]r3[ 0][ 0][ 0] + 
				([0]r3[ 0][ 0][-1] + 
				 [0]r3[ 0][ 0][ 1] + 
				 [0]r3[ 0][-1][ 0] + 
				 [0]r3[-1][ 1][ 0] + 
				 [0]r3[ 1][ 0][ 0] + 
				 [0]r3[0][0][ 0])*nu;
		}


		iterate 3 { // Hardcoded GAUSSIAN_STEP

			stencil gauss_blur {
				[1:dim2-2][1:dim1-2][1:dim0-2] : filter_7(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0:dim1-1][0     ] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0:dim1-1][dim0-1  ] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0     ][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][dim1-1  ][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0     ][0:dim1-1][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[dim2-1  ][0:dim1-1][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
			}
		} 
#pragma sdsl end

		// Post-scale
		for (i = 0; i < dim0*dim1*dim2; ++i) {
			V1[i] *= PostScale;
			V2[i] *= PostScale;
			V3[i] *= PostScale;
		}

		Ucal_CPU(V1, V2,V3,U1, U2, U3, S, T, interpT,i);
	}

#endif
	gettimeofday(&tim2, NULL);
	t2 = tim2.tv_sec+(tim2.tv_usec/1000000.0);
	printf("TIMESTAMP:reg sdsl for %d finishs at %lf diff %lf\r\n",imageID,t2,t2-t1);
	printf("SDSL kernel finished working on image %d \r\n",imageID);
  printf("Finished SDSL registration...\n");
	return T;
}

#ifdef STANDALONE_DRIVER
#include <cstdlib>
#include <ctime>
int main() {
  img = (float*)malloc(ImageN*ImageN*ImageP*sizeof(float));
  
  // Populate image with random values
  srand(time(NULL));
  for (int i = 0; i < ImageN*ImageN*ImageP; ++i) {
    img[i] = (float)rand() / (float)(RAND_MAX+1.0f) * 10.0f;
  }
  
  float *denoised_img = denoise_sdsl(0,img);
  float* reg_img = reg_sdsl(0,denoised_img);
  int seg_out = seg_sdsl(0,reg_img);
  
  return seg_out;
}
#endif
