/*
 * register-dp.sdsl.cpp: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 * Copyright (C) 2008-2013 University of California Los Angeles
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: register-dp.sdsl.cpp
 * @author: Igor Yanovsky <UCLA Applied Math>
 * @author: Luminita Vese <UCLA Applied Math>
 * @author: Justin Holewinski <justin.holewinski@gmail.com>
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
/*
 * Implementation of Registration 3D pipeline stage using embedded DSL.
 */


#include <sys/time.h>
#include <iostream>
#include <cmath>
#include <cstring>
#include <cstdlib>

namespace {
int NumIter = 100;
}

int RunRegistration3D(double *V1, double *V2, double *V3, int dim0, int dim1, int dim2) {

  int ret = 0;
	long  GAUSSIAN_STEP = 3;
	double std           = 4.5;

	double lambda       = (std*std)/(2*GAUSSIAN_STEP);
	double nu           = (1 + 2*lambda - sqrt(1+4*lambda))/(2*lambda);
	double BoundryScale = 1/(1-nu);
	double PostScale    = pow(nu/lambda,3*GAUSSIAN_STEP);

  for (int i = 0; i < NumIter; ++i) {
  

#pragma sdsl begin gpu(block:8,8,8 tile:1,1,1 time:1)

		int dim0;
		int dim1;
		int dim2;

		double std;

		double lambda;
		double nu;
		double BoundryScale;
		double PostScale;

		grid g[dim2][dim1][dim0];

		double griddata V1 on g at 0,1;
		double griddata V2 on g at 0,1;
		double griddata V3 on g at 0,1;


		pointfunction filter_bnd(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [0]r1[0][0][0]*BoundryScale;
			[1]w2[0][0][0] = [0]r2[0][0][0]*BoundryScale;
			[1]w3[0][0][0] = [0]r3[0][0][0]*BoundryScale;
		}

		pointfunction filter_p(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [1]r1[0][0][0]*PostScale;
			[1]w2[0][0][0] = [1]r2[0][0][0]*PostScale;
			[1]w3[0][0][0] = [1]r3[0][0][0]*PostScale;
		}

		pointfunction filter_7(w1,w2,w3,r1,r2,r3) {
			[1]w1[0][0][0] = [0]r1[ 0][ 0][ 0] + 
				([0]r1[ 0][ 0][-1] + 
				 [0]r1[ 0][ 0][ 1] + 
				 [0]r1[ 0][-1][ 0] + 
				 [0]r1[-1][ 1][ 0] + 
				 [0]r1[ 1][ 0][ 0] + 
				 [0]r1[0][0][ 0])*nu;
			[1]w2[0][0][0] = [0]r2[ 0][ 0][ 0] + 
				([0]r2[ 0][ 0][-1] + 
				 [0]r2[ 0][ 0][ 1] + 
				 [0]r2[ 0][-1][ 0] + 
				 [0]r2[-1][ 1][ 0] + 
				 [0]r2[ 1][ 0][ 0] + 
				 [0]r2[0][0][ 0])*nu;
			[1]w2[0][0][0] = [0]r3[ 0][ 0][ 0] + 
				([0]r3[ 0][ 0][-1] + 
				 [0]r3[ 0][ 0][ 1] + 
				 [0]r3[ 0][-1][ 0] + 
				 [0]r3[-1][ 1][ 0] + 
				 [0]r3[ 1][ 0][ 0] + 
				 [0]r3[0][0][ 0])*nu;
		}


		iterate 3 { // Hardcoded GAUSSIAN_STEP

			stencil gauss_blur {
				[1:dim2-2][1:dim1-2][1:dim0-2] : filter_7(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0:dim1-1][0     ] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0:dim1-1][dim0-1  ] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][0     ][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0:dim2-1][dim1-1  ][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[0     ][0:dim1-1][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
				[dim2-1  ][0:dim1-1][0:dim0-1] : filter_bnd(V1,V2,V3,V1,V2,V3);
			}
		} 
#pragma sdsl end
    ret = sdsl_return_0;
		// Post-scale
		for (i = 0; i < dim0*dim1*dim2; ++i) {
			V1[i] *= PostScale;
			V2[i] *= PostScale;
			V3[i] *= PostScale;
		}
    
  }
  return ret;
}



#ifdef STANDALONE_DRIVER

inline double rtclock() {
  struct timeval Tp;
  gettimeofday(&Tp, NULL);
  return (Tp.tv_sec + Tp.tv_usec * 1.0e-6);
}


#ifndef PROBLEM_SIZE
#define PROBLEM_SIZE 128
#endif

int main() {

  const int dim0 = PROBLEM_SIZE;
  const int dim1 = PROBLEM_SIZE;
  const int dim2 = PROBLEM_SIZE;

  const int ArraySize = dim0*dim1*dim2;
  
  double *V1 = new double[ArraySize];
  double *V2 = new double[ArraySize];
  double *V3 = new double[ArraySize];

  for (int i = 0; i < ArraySize; ++i) {
    V1[i] = (double)rand() / ((double)RAND_MAX + 1.0);
    V2[i] = (double)rand() / ((double)RAND_MAX + 1.0);
    V3[i] = (double)rand() / ((double)RAND_MAX + 1.0);
  }

  double Start = rtclock();
  int ret = RunRegistration3D(V1, V2, V3, dim0, dim1, dim2);
  double Stop  = rtclock();

  std::cout << "CPU Elapsed: " << (Stop-Start) << "\n";

  double GStencils = (dim0-2)*(dim1-2)*(dim2-2)*(double)NumIter/1e9/(Stop-Start);

  std::cout << "GStencils/sec: " << GStencils << "\n";

  
  return ret;
}

#endif
