/*
 * segment-dp.sdsl.cpp: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 * Copyright (C) 2001-2013 University of California Los Angeles
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: segment-dp.sdsl.cpp
 * @author: Tony Chan <Member, IEEE>
 * @author: Igor Yanovsky <UCLA Applied Math>
 * @author: Luminita Vese <UCLA Applied Math>
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 * @author: Justin Holewinski <justin.holewinski@gmail.com>
 */
/*
 * Implementation of Segmentation 3D pipeline stage using embedded DSL.
 */

#include <sys/time.h>
#include <iostream>
#include <cmath>
#include <cstring>
#include <cstdlib>

#ifndef TIME_STEPS
#define TIME_STEPS 50
#endif

#ifndef PROBLEM_SIZE
#define PROBLEM_SIZE 128
#endif


// WARNING: Changes in IterPerKernel here must also be manually changed in the
//          SDSL code
namespace {
const int MaxIter       = TIME_STEPS;
const int IterPerKernel = 10;
}

#define REF3D(i,j,k) ((i)*dim0*dim1+(j)*dim0+(k))

#define		MDP(a, b)	(2.0 * a / (b * b))
#define		MU	(0.18 * 255 * 255)
#define   DX  1.0
#define   DY  1.0
#define   DZ  1.0
#define		SUM	((MDP(MU, DX)) + (MDP(MU, DY)) + (MDP(MU, DZ)))
#define		DT1	((SUM > 1.0e-15) ? (0.9 / SUM) : (100.0)) /* Compute the Courant number for Active Contours CV */
#define		DT2	(10e7 * DT1)


int RunSegmentation3D(double *u0, double *phi, int dim0, int dim1, int dim2) {

  double num1, num2, den1, den2, c1, c2;

  int ret = 0;
  num1 = 0.0;
  num2 = 0.0;
  den1 = 0.0;
  den2 = 0.0;

  for (int i = 0; i < MaxIter; i += IterPerKernel) {

    for(int i = 0; i < dim2; i++) {
      for(int j = 0; j < dim1; j++) {
        for(int k = 0; k < dim0; k++) {
          num1 = (phi[REF3D(i,j,k)] < 0) ? num1 + u0[REF3D(i,j,k)] : (num1);
          den1 = (phi[REF3D(i,j,k)] < 0) ? den1 + 1 : (den1);
          num2 = (phi[REF3D(i,j,k)] > 0) ? num2 + u0[REF3D(i,j,k)] : (num2);
          den2 = (phi[REF3D(i,j,k)] > 0) ? den2 + 1 : (den2);
        }
      }
    }

    c1 = num1 / den1;
    c2 = num2 / den2;

    double DT = DT2;

    #pragma sdsl begin gpu(block:8,8,8 tile:1,1,1 time:1) // gpu(...) optional
    int dim0;
    int dim1;
    int dim2;
    
    double c1;
    double c2;
    double DT;
    
    double DX = 1.0;
    double DY = 1.0;
    double DZ = 1.0;
    
    double DX2 = DX * 2.0;
    double DY2 = DY * 2.0;
    double DZ2 = DZ * 2.0;
    
    double MU = 0.18 * 255.0 * 255.0;
    double LAMBDA1 = 1.0;
    double LAMBDA2 = 1.0;
    double EPSILON = 10.0E-7;
    double EPSILON_SQRD = EPSILON*EPSILON;
    
    
    grid g[dim2][dim1][dim0];
    
    double griddata u0  on g at 0;
    double griddata phi on g at 0,1;
    
    pointfunction calc_phi(p,u) {
      double Dx_p = ([0]p[1][0][0] - [0]p[ 0][ 0][ 0])/DX;
      double Dx_m = ([0]p[0][0][0] - [0]p[-1][ 0][ 0])/DX;
      double Dy_p = ([0]p[0][1][0] - [0]p[ 0][ 0][ 0])/DY;
      double Dy_m = ([0]p[0][0][0] - [0]p[ 0][-1][ 0])/DY;
      double Dz_p = ([0]p[0][0][1] - [0]p[ 0][ 0][ 0])/DZ;
      double Dz_m = ([0]p[0][0][0] - [0]p[ 0][ 0][-1])/DZ;
    
      double Dx_0 = ([0]p[1][0][0] - [0]p[-1][ 0][ 0])/DX2;
      double Dy_0 = ([0]p[0][1][0] - [0]p[ 0][-1][ 0])/DY2;
      double Dz_0 = ([0]p[0][0][1] - [0]p[ 0][ 0][-1])/DZ2;
    
      double C1x  =  rsqrt( Dx_p*Dx_p + Dy_0*Dy_0 + Dz_0*Dz_0  + EPSILON_SQRD);
      double C2x  =  rsqrt( Dx_m*Dx_m + Dy_0*Dy_0 + Dz_0*Dz_0  + EPSILON_SQRD);
      double C3y  =  rsqrt( Dx_0*Dx_0 + Dy_p*Dy_p + Dz_0*Dz_0  + EPSILON_SQRD);
      double C4y  =  rsqrt( Dx_0*Dx_0 + Dy_m*Dy_m + Dz_0*Dz_0  + EPSILON_SQRD);
      double C5z  =  rsqrt( Dx_0*Dx_0 + Dy_0*Dy_0 + Dz_p*Dz_p  + EPSILON_SQRD);
      double C6z  =  rsqrt( Dx_0*Dx_0 + Dy_0*Dy_0 + Dz_m*Dz_m  + EPSILON_SQRD);
    
      double Grad = sqrt(Dx_0*Dx_0 + Dy_0*Dy_0 + Dz_0*Dz_0);
    
      double MM  =  (DT/(DX*DY)) * Grad * MU;
      double CC  =  1 + MM*(C1x + C2x + C3y + C4y + C5z + C6z);
    
      double C1x_2x  =  C1x*[0]p[1][0][0] + C2x*[0]p[-1][ 0][ 0];
      double C3y_4y  =  C3y*[0]p[0][1][0] + C4y*[0]p[ 0][-1][ 0];
      double C5z_6z  =  C5z*[0]p[0][0][1] + C6z*[0]p[ 0][ 0][-1];
    
      [1]p[0][0][0] = (1.0 / CC) * ( [0]p[0][0][0] + MM*( C1x_2x + C3y_4y + C5z_6z )
                                  + (DT*Grad)* ( (LAMBDA1*([0]u[0][0][0] - c1)*([0]u[0][0][0] - c1))
                                  -              (LAMBDA2*([0]u[0][0][0] - c2)*([0]u[0][0][0] - c2))));
    }
    
    
    iterate 10 { // IterPerKernel hardcoded
      stencil segi {
        [1:dim2-2][1:dim1-2][1:dim0-2] : calc_phi(phi, u0);
        [0:dim2-1][0:dim1-1][0       ] : [1]phi[0][0][0]   = [0]phi[0][0][1 ];
        [0:dim2-1][0:dim1-1][dim0-1  ] : [1]phi[0][0][0] = [0]phi[0][0][-1];
        
        [0:dim2-1][0       ][0:dim0-1] : [1]phi[0][0][0] = [0]phi[0][1 ][0];
        [0:dim2-1][dim1-1  ][0:dim0-1] : [1]phi[0][0][0] = [0]phi[0][-1][0];
        
        [0       ][0:dim1-1][0:dim0-1] : [1]phi[0][0][0] = [0]phi[1 ][0][0];
        [dim2-1  ][0:dim1-1][0:dim0-1] : [1]phi[0][0][0] = [0]phi[-1][0][0];
    
        [0     ][0     ][0     ] : [1]phi[0][0][0] = [0]phi[ 1][ 1][ 1];
        [0     ][0     ][dim0-1] : [1]phi[0][0][0] = [0]phi[ 1][ 1][-1];
        [0     ][dim1-1][0     ] : [1]phi[0][0][0] = [0]phi[ 1][-1][ 1];
        [0     ][dim1-1][dim0-1] : [1]phi[0][0][0] = [0]phi[ 1][-1][-1];
        [dim2-1][0     ][0     ] : [1]phi[0][0][0] = [0]phi[-1][ 1][ 1];
        [dim2-1][0     ][dim0-1] : [1]phi[0][0][0] = [0]phi[-1][ 1][-1];
        [dim2-1][dim1-1][0     ] : [1]phi[0][0][0] = [0]phi[-1][-1][ 1];
        [dim2-1][dim1-1][dim0-1] : [1]phi[0][0][0] = [0]phi[-1][-1][-1];
    
      }
    } 
    #pragma sdsl end
      ret = sdsl_return_0;
    }
  return ret;
}


#ifdef STANDALONE_DRIVER

#ifndef TIME_STEPS
#define TIME_STEPS 50
#endif

#ifndef PROBLEM_SIZE
#define PROBLEM_SIZE 200
#endif

inline double rtclock() {
  struct timeval Tp;
  gettimeofday(&Tp, NULL);
  return (Tp.tv_sec + Tp.tv_usec * 1.0e-6);
}


#define		XCENT	((M - 1) / 2.0)
#define		YCENT	((N - 1) / 2.0)
#define		ZCENT	((P - 1) / 2.0)
#define		RR	((MIN(MIN(M , N), P)) / 3.0)
#define		DX 	1.0
#define		DY	1.0
#define		DZ	1.0
#define		DX2	(DX * 2.0)
#define		DY2	(DY * 2.0)
#define		DZ2	(DZ * 2.0)
#define		XMIN	0.0
#define		XMAX	((M - 1) * DX)
#define		YMIN	0.0
#define		YMAX	((N - 1) * DY)
#define		ZMIN	0.0
#define		ZMAX	((P - 1) * DZ)

/* AC parameter definition */
#define		MU	(0.18 * 255 * 255)
#define		LAMBDA1	1.0
#define		LAMBDA2	1.0
#define		METHOD	'e'
#define		GRAD	'g'

/* runData parameter definition */
#define		EPSILON		10e-7
#define		EPSILON_SQRD	(EPSILON * EPSILON)


typedef double Array3[][PROBLEM_SIZE][PROBLEM_SIZE];
typedef double (*Array3Ptr)[PROBLEM_SIZE][PROBLEM_SIZE];

void reference(Array3 phi_ref, Array3 u0_ref, Array3 phi_new) {

  int       i, j, k, kk;
  double    num1_ref, num2_ref, den1_ref, den2_ref;
  const int M  = PROBLEM_SIZE;
  const int N  = PROBLEM_SIZE;
  const int P  = PROBLEM_SIZE;
  double    c1_ref, c2_ref;
  double    DT = DT2;
  
  memcpy(phi_new, phi_ref, sizeof(double)*M*N*P);
  
  for(kk = 1; kk <= TIME_STEPS; kk++) {
    // advanceTimeStepIMPLICIT( phi, u0, gridData, runData, AC ) inlined 

    /* getMeanIntensities( u0, phi, AC ) inlined */
    num1_ref = 0.0;
    num2_ref = 0.0;
    den1_ref = 0.0;
    den2_ref = 0.0;

    for(i=0; i < M; i++) {
      for(j=0; j < N; j++) {
        for(k=0; k < P; k++) {
          if(phi_ref[i][j][k] < 0) {
            num1_ref = num1_ref + u0_ref[i][j][k];
            den1_ref = den1_ref + 1;
          }
          else if(phi_ref[i][j][k] > 0) {
            num2_ref = num2_ref + u0_ref[i][j][k];
            den2_ref = den2_ref + 1;
          }
        }
      }
    }

    c1_ref = num1_ref / den1_ref;
    c2_ref = num2_ref / den2_ref;
    /* end of inlining getMeanIntensities */

    for(i = 1; i < M - 1; i++) {
      for(j = 1; j < N - 1; j++) {
        for(k = 1; k < P - 1; k++) {
          double Dx_p_ref = (phi_ref[i+1][j][k] - phi_ref[i][j][k])/DX;
          double Dx_m_ref = (phi_ref[i][j][k] - phi_ref[i-1][j][k])/DX;
          double Dy_p_ref = (phi_ref[i][j+1][k] - phi_ref[i][j][k])/DY;
          double Dy_m_ref = (phi_ref[i][j][k] - phi_ref[i][j-1][k])/DY;
          double Dz_p_ref = (phi_ref[i][j][k+1] - phi_ref[i][j][k])/DZ;
          double Dz_m_ref = (phi_ref[i][j][k] - phi_ref[i][j][k-1])/DZ;

          double Dx_0_ref = (phi_ref[i+1][j][k] - phi_ref[i-1][j][k])/DX2;
          double Dy_0_ref = (phi_ref[i][j+1][k] - phi_ref[i][j-1][k])/DY2;
          double Dz_0_ref = (phi_ref[i][j][k+1] - phi_ref[i][j][k-1])/DZ2;

          double C1x_ref = 1.0 / sqrt( Dx_p_ref*Dx_p_ref + Dy_0_ref*Dy_0_ref + Dz_0_ref*Dz_0_ref  + EPSILON_SQRD);
          double C2x_ref = 1.0 / sqrt( Dx_m_ref*Dx_m_ref + Dy_0_ref*Dy_0_ref + Dz_0_ref*Dz_0_ref  + EPSILON_SQRD);
          double C3y_ref = 1.0 / sqrt( Dx_0_ref*Dx_0_ref + Dy_p_ref*Dy_p_ref + Dz_0_ref*Dz_0_ref  + EPSILON_SQRD);
          double C4y_ref = 1.0 / sqrt( Dx_0_ref*Dx_0_ref + Dy_m_ref*Dy_m_ref + Dz_0_ref*Dz_0_ref  + EPSILON_SQRD);
          double C5z_ref = 1.0 / sqrt( Dx_0_ref*Dx_0_ref + Dy_0_ref*Dy_0_ref + Dz_p_ref*Dz_p_ref  + EPSILON_SQRD);
          double C6z_ref = 1.0 / sqrt( Dx_0_ref*Dx_0_ref + Dy_0_ref*Dy_0_ref + Dz_m_ref*Dz_m_ref  + EPSILON_SQRD);

          double Grad_ref = sqrt(Dx_0_ref*Dx_0_ref + Dy_0_ref*Dy_0_ref + Dz_0_ref*Dz_0_ref);

          double MM_ref = (DT/(DX*DY)) * Grad_ref * MU;
          double CC_ref = 1 + MM_ref*(C1x_ref + C2x_ref + C3y_ref + C4y_ref + C5z_ref + C6z_ref);

          double C1x_2x_ref = C1x_ref*phi_ref[i+1][j][k] + C2x_ref*phi_ref[i-1][j][k];
          double C3y_4y_ref = C3y_ref*phi_ref[i][j+1][k] + C4y_ref*phi_ref[i][j-1][k];
          double C5z_6z_ref = C5z_ref*phi_ref[i][j][k+1] + C6z_ref*phi_ref[i][j][k-1];

          phi_new[i][j][k] = (1.0 / CC_ref) * ( phi_ref[i][j][k] + MM_ref*( C1x_2x_ref + C3y_4y_ref + C5z_6z_ref )
                                                + (DT*Grad_ref)* ( (LAMBDA1*(u0_ref[i][j][k] - c1_ref)*(u0_ref[i][j][k] - c1_ref))
                                                                   - (LAMBDA2*(u0_ref[i][j][k] - c2_ref)*(u0_ref[i][j][k] - c2_ref))));
        }
      }
    }

    /* NeumannBC( phi ) inlined */
    for(j=0; j < N; j++) {
      for(k=0; k < P; k++) {
        phi_new[0][j][k]   = phi_ref[1][j][k];
        phi_new[M-1][j][k] = phi_ref[M-2][j][k];
      }
    }

    for(i=0; i < M; i++) {
      for(k=0; k < P; k++) {
        phi_new[i][0][k]   = phi_ref[i][1][k];
        phi_new[i][N-1][k] = phi_ref[i][N-2][k];
      }
    }
    
    for(i=0; i < M; i++) {
      for(j=0; j < N; j++) {
        phi_new[i][j][0]     = phi_ref[i][j][1];
        phi_new[i][j][P-1] = phi_ref[i][j][P-2];
      }
      }

    phi_new[0][0][0]     = phi_ref[1][1][1];
    phi_new[M-1][0][0]     = phi_ref[M-2][1][1];
    phi_new[0][N-1][0]     = phi_ref[1][N-2][1];
    phi_new[0][0][P-1]     = phi_ref[1][1][P-2];
    phi_new[M-1][P-1][0]   = phi_ref[M-2][N-2][1];
    phi_new[M-1][0][P-1]   = phi_ref[M-2][1][P-2];
    phi_new[0][N-1][P-1]   = phi_ref[1][N-2][P-2];
    phi_new[M-1][N-1][P-1] = phi_ref[M-2][N-2][P-2];
    /* end of inlining NeumannBC */

    /* end of inlining advanceTimeStepEXPLICIT */

    memcpy(phi_ref, phi_new, sizeof(double)*M*N*P);
  }                             /* end of for (timesteps) */
}

template <typename T>
bool CompareResult(T *Result, T *Reference, size_t Size) {
  
  T ErrorNorm = 0.0;
  T RefNorm   = 0.0;

  for (unsigned i = 0; i < Size; ++i) {
    float Diff  = Result[i] - Reference[i];
    ErrorNorm  += Diff*Diff;
    RefNorm    += Reference[i]*Reference[i];
  }

  ErrorNorm = std::sqrt(ErrorNorm);
  RefNorm   = std::sqrt(RefNorm);

  std::cout << "Error Norm:  " << ErrorNorm << "\n";
  std::cout << "Ref Norm:    " << RefNorm << "\n";
 
  if(std::abs(RefNorm) < 1e-7) {
    std::cout << "FAIL!\n";
    return false;
  }
  else if((ErrorNorm / RefNorm) > 1e-2) {
    std::cout << "FAIL!\n";
    return false;
  }
  else {
    std::cout << "OK\n";
    return true;
  }
}

int main() {

  const int dim0     = PROBLEM_SIZE;
  const int dim1     = PROBLEM_SIZE;
  const int dim2     = PROBLEM_SIZE;
  const int TimeSteps = 100;
  
  double *u0 = new double[dim0*dim1*dim2];
  double *phi = new double[dim0*dim1*dim2];

  //srand(time(NULL));
  srand(4242);
  
  for (int i = 0; i < dim0*dim1*dim2; ++i) {
    u0[i]  = (double)rand() / (double)(RAND_MAX+1.0f) - 0.5;
    phi[i] = (double)rand() / (double)(RAND_MAX+1.0f) - 0.5;
  }

#ifdef CHECK_REF
  double *phi_ref = new double[dim0*dim1*dim2];
  double *u0_ref  = new double[dim0*dim1*dim2];
  double *phi_new = new double[dim0*dim1*dim2];;
  
  memcpy(phi_ref, phi, sizeof(double)*PROBLEM_SIZE*PROBLEM_SIZE*PROBLEM_SIZE);
  memcpy(u0_ref, u0, sizeof(double)*PROBLEM_SIZE*PROBLEM_SIZE*PROBLEM_SIZE);
  memcpy(phi_new, phi, sizeof(double)*PROBLEM_SIZE*PROBLEM_SIZE*PROBLEM_SIZE);
  
  double RefStart = rtclock();
  reference((Array3Ptr)phi_ref, (Array3Ptr)u0_ref, (Array3Ptr)phi_new);
  double RefStop  = rtclock();
#endif

  double Start = rtclock();
  
  int Ret = RunSegmentation3D(u0, phi, dim0, dim1, dim2);

  double Stop = rtclock();

  std::cout << "CPU Elapsed: " << (Stop-Start) << "\n";

  double GStencils = (dim0-2)*(dim1-2)*(dim2-2)*(double)TimeSteps/1e9/(Stop-Start);

  std::cout << "GStencils/sec: " << GStencils << "\n";

#ifdef CHECK_REF

  double Speedup = (RefStop-RefStart) / (Stop-Start);
  std::cout << "Speedup over CPU ref: " << Speedup << "x\n";

  double CPUGStencils = (dim0-2)*(dim1-2)*(dim2-2)*(double)TimeSteps/1e9/(RefStop-RefStart);

  std::cout << "CPU GStencils/sec: " << CPUGStencils << "\n";

  CompareResult(phi, (double*)phi_ref, PROBLEM_SIZE*PROBLEM_SIZE*PROBLEM_SIZE);
#endif
  
  delete [] u0;
  delete [] phi;

  return Ret;
}

#endif
