/*
 * j1d3-dp.sdsl.c: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: j1d3-dp.sdsl.c
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <assert.h>

#include "timing.h"

const double ONETHIRD = (0.33333333333);


// Problem parameters
#ifndef W
#define W (8000*8000)
#endif

#ifndef T
#define T (100)
#endif

double aref[2][W];
double a[W];



/** Main program */
int main(int argc, char *argv[]) {
  int i, ii, t, tt;
  double refElapsed;
  double sdslElapsed;
  double refGFLOPS;
  double sdslGFLOPS;
  double *inref, *outref;
  
  
  // Initialize arrays
  for (i = 0; i < W; i++) {
      aref[0][i] = aref[1][i] = cos(i);
      a[i] = cos(i);
  }
  
  // Compute reference
  #ifndef NOREF
  flush_llc();
  init_timer();
  start_timer();
  for (t = 0; t < T; t++) {
    inref  = aref[t & 1];
    outref = aref[(t + 1) & 1];
    #pragma ivdep
    #pragma vector always
    for (i = 1; i < W - 1; i++) {
        outref[i] = ONETHIRD*(inref[i-1] + inref[i] + inref[i+1]);
    }
  }
  refElapsed = stop_timer();
  refGFLOPS = ((double)3)*((double)(W - 2))*((double)T)/refElapsed/1.0e9;
  printf("REF computed in %.6f sec (%.2f GFLOPS)\n",refElapsed,refGFLOPS);
  #endif
  
  
  // Compute SDSL
  const int dim0 = W;
  
  flush_llc();
  reset_timer();
  start_timer();
  #pragma sdsl begin gpu(block:8 tile:1 time:1)
  int dim0;
  
  double ONETHIRD;
  
  grid g[dim0];
  
  double griddata a on g at 0,1;
  
  pointfunction compute(x) {
        [1]x[0] = ONETHIRD*([0]x[-1] + [0]x[0] + [0]x[1]);
  }
  
  iterate 100 {
    stencil main {
      [1:dim0-2] : compute(a);
    }
  }
  #pragma sdsl end
  sdslElapsed = stop_timer();
  sdslGFLOPS = ((double)3)*((double)(W - 2))*((double)T)/sdslElapsed/1.0e9;
  #ifndef NOREF
  printf("SDSL computed in %.6f sec (%.2f GFLOPS)\n",sdslElapsed,sdslGFLOPS);
  printf("Speedup is %.2fx\n",refElapsed/sdslElapsed);
  
  // Check correctness
  outref = aref[T & 1];float maxDiff = 0.0, sumDiff = 0.0, diff, meanDiff;
  int numDiff = 0;
  for (i = 0; i < W; i++) {
      diff = fabs(outref[i] - a[i]);
      if (diff != 0.0f) {
        #ifdef DEBUG
        printf("Diff at a[%d] = %.6f\n",i,diff);
        #endif
        numDiff++;
      }
      sumDiff += diff;
      if (diff > maxDiff) {
        maxDiff = diff;
      }
  }
  meanDiff = sumDiff/((float)(W));
  
  printf("Num diff  = %d\n",numDiff);
  printf("Sum diff  = %.8f\n",sumDiff);
  printf("Mean diff = %.8f\n",meanDiff);
  printf("Max diff  = %.8f\n",maxDiff);
  #else           
  printf("%.2f\n",sdslGFLOPS);
  srand(time(NULL));
  int rand_i = rand() % W;
  
  int numDiff = (int)a[rand_i];
  #endif
  
  return sdsl_return_0;
}
