/*
 * j3d7-sp.sdsl.c: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: j3d7-sp.sdsl.c
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <assert.h>

#include "timing.h"

const float ONESEVENTH = (0.142857143f);

// Problem parameters
#ifndef D
#define D (400)
#endif

#ifndef H
#define H (400)
#endif

#ifndef W
#define W (400)
#endif
                                                                    
#ifndef T
#define T (100)
#endif

  
float aref[2][D][H][W];
float a[D][H][W];


/** Main program */
int main(int argc, char* argv[]) {
  int i, j, k, ii, jj, kk, t, tt;
  double refElapsed, sdslElapsed;
  double refGFLOPS, sdslGFLOPS;
  float (*inref)[H][W], (*outref)[H][W];
  float *asdsl = &a[0][0][0];
  
  // Initialize arrays
  for (i = 0; i < D; i++) {
    for (j = 0; j < H; j++) {
      for (k = 0; k < W; k++) {
        aref[0][i][j][k] = aref[1][i][j][k] = sin(i)*cos(j)*tan(k);
        a[i][j][k] = sin(i)*cos(j)*tan(k);
      }
    }
  }
  
  // Compute reference
  #ifndef NOREF
  flush_llc();
  init_timer();
  start_timer();
  for (t = 0; t < T; t++) {
    inref  = aref[t & 1];
    outref = aref[(t + 1) & 1];
    for (i = 1; i < D - 1; i++) {
      for (j = 1; j < H - 1; j++) {
        #pragma ivdep
        #pragma vector always
        for (k = 1; k < W - 1; k++) {
          outref[i][j][k] = ONESEVENTH*(inref[i-1][j+0][k+0] + inref[i+0][j-1][k+0] + 
                                        inref[i+0][j+0][k-1] + inref[i+0][j+0][k+0] + inref[i+0][j+0][k+1] + 
                                        inref[i+0][j+1][k+0] + inref[i+1][j+0][k+0]);
        }
      }
    }
  }
  refElapsed = stop_timer();
  refGFLOPS = ((double)7)*((double)(D - 2)*(H - 2)*(W - 2))*((double)T)/refElapsed/1.0e9;
  printf("REF computed in %.6f sec (%.2f GFLOPS)\n",refElapsed,refGFLOPS);
  #endif
  
  
  // Compute SDSL
  const int dim0 = W;
  const int dim1 = H;
  const int dim2 = D;
  
  flush_llc();
  reset_timer();
  start_timer();
  #pragma sdsl begin gpu(block:8,8,8 tile:1,1,1 time:1)
  int dim0;
  int dim1;
  int dim2;
  
  float ONESEVENTH;
  
  grid g[dim2][dim1][dim0];
  
  float griddata asdsl on g at 0,1;
  
  pointfunction compute(x) {
          [1]x[0][0][0] = ONESEVENTH*([0]x[-1][0][ 0] + [0]x[0][-1][0] + 
                                      [0]x[ 0][0][-1] + [0]x[0][ 0][0] + [0]x[0][0][1] + 
                                      [0]x[ 0][1][ 0] + [0]x[1][ 0][0]);
  }
  
  iterate 100 {
    stencil main {
      [1:dim2-2][1:dim1-2][1:dim0-2] : compute(asdsl);
    }
  }
  #pragma sdsl end
  sdslElapsed = stop_timer();
  sdslGFLOPS = ((double)7)*((double)(D - 2)*(H - 2)*(W - 2))*((double)T)/sdslElapsed/1.0e9;
  #ifndef NOREF
  printf("SDSL computed in %.6f sec (%.2f GFLOPS)\n",sdslElapsed,sdslGFLOPS);
  printf("Speedup is %.2fx\n",refElapsed/sdslElapsed);
  
  // Check correctness
  outref = aref[T & 1];float maxDiff = 0.0, sumDiff = 0.0, diff, meanDiff;
  int numDiff = 0;
  for (i = 0; i < D; i++) {
    for (j = 0; j < H; j++) {
      for (k = 0; k < W; k++) {
        diff = fabs(outref[i][j][k] - a[i][j][k]);
        if (diff != 0.0f) {
          #ifdef DEBUG
          printf("Diff at a[%d][%d][%d] = %.6f\n",i,j,k,diff);
          #endif
          numDiff++;
        }
        sumDiff += diff;
        if (diff > maxDiff) {
          maxDiff = diff;
        }
      }
    }
  }
  meanDiff = sumDiff/((float)(H*W));
  
  printf("Num diff  = %d\n",numDiff);
  printf("Sum diff  = %.8f\n",sumDiff);
  printf("Mean diff = %.8f\n",meanDiff);
  printf("Max diff  = %.8f\n",maxDiff);
  #else           
  printf("%.2f\n",sdslGFLOPS);
  srand(time(NULL));
  int rand_i = rand() % H;
  int rand_j = rand() % W;
  
  int numDiff = (int)a[rand_i][rand_j][rand_j];
  #endif
  
  return sdsl_return_0;
}
