/*
 * lapl-2d-dp.sdsl.c: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: lapl-2d-dp.sdsl.c
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <assert.h>

#include "timing.h"

                     
// Problem parameters
#ifndef H
#define H (8000)
#endif

#ifndef W
#define W (8000)
#endif

#ifndef T
#define T (100)
#endif


double aref[2][H][W];
double a[H][W];



const double ALPHA = (6.0);
const double BETA = (0.9);


/** Main program */
int main(int argc, char *argv[]) {
  int i, j, ii, jj, t, tt;
  double refElapsed, sdslElapsed;
  double refGFLOPS, sdslGFLOPS;
  double (*inref)[W], (*outref)[W];
  double *asdsl = &a[0][0];
  
  
  // Initialize arrays
  for (i = 0; i < H; i++) {
    for (j = 0; j < W; j++) {
      aref[0][i][j] = aref[1][i][j] = sin(i)*cos(j);
      a[i][j] = sin(i)*cos(j);
    }
  }
  
  // Compute reference
  #ifndef NOREF
  flush_llc();
  init_timer();
  start_timer();
  for (t = 0; t < T; t++) {
    inref  = aref[t & 1];
    outref = aref[(t + 1) & 1];
    for (i = 1; i < H - 1; i++) {
      #pragma ivdep
      #pragma vector always
      for (j = 1; j < W - 1; j++) {
         outref[i][j] = ALPHA*inref[i][j] - BETA*(inref[i-1][j]  + inref[i][j-1] + inref[i][j+1] + inref[i+1][j]);
      }
    }
  }
  refElapsed = stop_timer();
  refGFLOPS = ((double)6)*((double)(H - 2)*(W - 2))*((double)T)/refElapsed/1.0e9;
  printf("REF computed in %.6f sec (%.2f GFLOPS)\n",refElapsed,refGFLOPS);
  #endif
  
  
  // Compute SDSL
  const int dim0 = W;
  const int dim1 = H;
  
  flush_llc();
  reset_timer();
  start_timer();
  #pragma sdsl begin gpu(block:8,8 tile:1,1 time:1)
  int dim0;
  int dim1;
  
  double ALPHA;
  double BETA;
  
  grid g[dim1][dim0];
  
  double griddata asdsl on g at 0,1;
  
  pointfunction compute(x) {
         [1]x[0][0] = ALPHA*[0]x[0][0] - BETA*([0]x[-1][0]  + [0]x[0][-1] + [0]x[0][1] + [0]x[1][0]);
  }
  
  iterate 100 {
    stencil main {
      [1:dim1-2][1:dim0-2] : compute(asdsl);
    }
  }
  #pragma sdsl end
  sdslElapsed = stop_timer();
  sdslGFLOPS = ((double)6)*((double)(H - 2)*(W - 2))*((double)T)/sdslElapsed/1.0e9;
  #ifndef NOREF
  printf("SDSL computed in %.6f sec (%.2f GFLOPS)\n",sdslElapsed,sdslGFLOPS);
  printf("Speedup is %.2fx\n",refElapsed/sdslElapsed);
  
  // Check correctness
  outref = aref[T & 1];float maxDiff = 0.0, sumDiff = 0.0, diff, meanDiff;
  int numDiff = 0;
  for (i = 0; i < H; i++) {
    for (j = 0; j < W; j++) {
      diff = fabs(outref[i][j] - a[i][j]);
      if (diff != 0.0f) {
        #ifdef DEBUG
        printf("Diff at a[%d][%d] = %.6f\n",i,j,diff);
        #endif
        numDiff++;
      }
      sumDiff += diff;
      if (diff > maxDiff) {
        maxDiff = diff;
      }
    }
  }
  meanDiff = sumDiff/((float)(H*W));
  
  printf("Num diff  = %d\n",numDiff);
  printf("Sum diff  = %.8f\n",sumDiff);
  printf("Mean diff = %.8f\n",meanDiff);
  printf("Max diff  = %.8f\n",maxDiff);
  #else           
  printf("%.2f\n",sdslGFLOPS);
  srand(time(NULL));
  int rand_i = rand() % H;
  int rand_j = rand() % W;
  
  int numDiff = (int)a[rand_i][rand_j];
  #endif
  
  return sdsl_return_0;
}
