/*
 * ASTParser.g: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

/*
 * @file: ASTParser.g
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */
tree grammar ASTParser;

options {
  tokenVocab=SDSL;
  backtrack=true;
  memoize=true;
  ASTLabelType=CommonTree;
}

@header {
package edu.osu.cse.hpcrl.sdsl.frontend;

import edu.osu.cse.hpcrl.sdsl.*;
import edu.osu.cse.hpcrl.sdsl.ir.*;
}

@members {
private SDSLProgram p = new SDSLProgram();
  
public SDSLProgram getSDSLProgram() {
  return p;
}

public static void pr(String s) {
	System.err.println(s);
}
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Parser rules
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 * Root of a SDSL program
 */
sdsl
: ^(SDSL_PROG decl_list);



////////////////////////////////////////////////////////////////////////////////
// Lists
////////////////////////////////////////////////////////////////////////////////
decl_list: decl*;

time_list returns [ArrayList<Expression> tList]
@init {
$tList = new ArrayList<Expression>();
}: 
(e1=expr {$tList.add($e1.e);}) (e2=expr {$tList.add($e2.e);})*;


grid_data_list returns [ArrayList<GridData> gdList]
@init {
$gdList = new ArrayList<GridData>();
}: 
(id1=ID {GridData gd = new GridData($id1.text); $gdList.add(gd);}) (id2=ID {GridData gd = new GridData($id2.text); $gdList.add(gd);})*;


arg_list returns [ArrayList<Expression> eList]
@init {
$eList = new ArrayList<Expression>();
}: 
(e1=expr {$eList.add($e1.e);}) (e2=expr {$eList.add($e2.e);})*;

iter_list returns [ArrayList<Computable> iList]
@init {
$iList = new ArrayList<Computable>();
}: c1=computable {iList.add($c1.c);} (c2=computable {iList.add($c2.c);})*;
computable returns [Computable c]: 
stencil_def {$c = $stencil_def.s;}
| reduce_def {$c = $reduce_def.r;};



////////////////////////////////////////////////////////////////////////////////
// Declarations and definitions
////////////////////////////////////////////////////////////////////////////////
decl: param_decl      {p.addParam($param_decl.pm);}
    | const_def       {p.addConst($const_def.c);}
    | grid_def        {p.addGrid($grid_def.g);}
    | grid_data_def   {p.addGridData($grid_data_def.gd);}
    | pf_def          {p.addPointFunction($pf_def.pf);}
    | iter_def        {p.setIteration($iter_def.i);};



////////////////////////////////////////////////////////////////////////////////
// Parameters / constants
////////////////////////////////////////////////////////////////////////////////
param_decl returns [Parameter pm]: ^(PARAM_DECL ID type) 
{ 
$pm = new Parameter($ID.text, $type.text); 
};

const_def returns [Constant c]: ^(CONST_DEF ID type expr) 
{ 
$c = new Constant($ID.text, $type.text, $expr.e); 
};

type: I_TYPE
    | L_TYPE
    | F_TYPE
    | D_TYPE;



////////////////////////////////////////////////////////////////////////////////
// Grid definition
////////////////////////////////////////////////////////////////////////////////
grid_def returns [Grid g]: ^(GRID_DEF ID const_region) 
{
$g = new Grid($ID.text,$const_region.r);
};



////////////////////////////////////////////////////////////////////////////////
// Grid Data definition
////////////////////////////////////////////////////////////////////////////////
grid_data_def returns [GridData gd]: ^(GRID_DATA_DEF type name=ID grid=ID (tl=time_list)?) 
{
String gdName = $name.text;
String type = $type.text;
Grid g = p.getGrid($grid.text);
ArrayList<Expression> tList = null;
if ($tl.text != null) {
  tList = $tl.tList;
}
else {
  tList = new ArrayList<Expression>();
  tList.add(new IntLiteral(0));
}
$gd = new GridData(gdName,type,g,tList);
};



////////////////////////////////////////////////////////////////////////////////
// Point Function definition
////////////////////////////////////////////////////////////////////////////////
pf_def returns [PointFunction pf]
@init {
ArrayList<PointFunctionStatement> pfsl = new ArrayList<PointFunctionStatement>();
}:
^(POINT_FUNCTION_DEF ID ^(PARAMETER_LIST grid_data_list?) ^(STMT_LIST (pf_stmt {pfsl.add($pf_stmt.pfs);})+))
{
ArrayList<GridData> gdl = null;
if ($grid_data_list.text != null) {
  gdl = $grid_data_list.gdList;
}
$pf = new PointFunction($ID.text, gdl, pfsl);
}
; 

////////////////////////////////////////////////////////////////////////////////
// Iteration definition
////////////////////////////////////////////////////////////////////////////////
iter_def returns [Iteration i]
@init {
$i = new Iteration();
}:
^(ITER_DEF i1=INT iter_list expr? i2=INT?) {
  $i.setMaxIterations(Integer.parseInt($i1.text));
  $i.setComputables($iter_list.iList);
  if ($expr.text != null) {
    $i.setConvergeCheck($expr.e);
  }
  if ($i2.text != null) {
    $i.setCheckEvery(Integer.parseInt($i2.text));
  }
};


////////////////////////////////////////////////////////////////////////////////
// Stencil definition
////////////////////////////////////////////////////////////////////////////////
stencil_def returns [Stencil s]
@init
{
ArrayList<StencilStatement> ssl = new ArrayList<StencilStatement>();
}
: ^(STENCIL_DEF ID ^(SUBCALC_LIST (stencil_stmt {ssl.add($stencil_stmt.ss);})+))
{
$s = new Stencil($ID.text,ssl);
};



////////////////////////////////////////////////////////////////////////////////
// Reduction definition
////////////////////////////////////////////////////////////////////////////////
reduce_def returns [Reduction r]
@init
{
ArrayList<ReductionStatement> rsl = new ArrayList<ReductionStatement>();
}: ^(REDUCTION_DEF ID assoc_op ^(SUBCALC_LIST (reduce_stmt {rsl.add($reduce_stmt.rs);})+))
{
$r = new Reduction($ID.text, $assoc_op.text, rsl);
};
	
assoc_op: PLUS
        | MINUS
        | MAX
        | MIN;



////////////////////////////////////////////////////////////////////////////////
// Region
////////////////////////////////////////////////////////////////////////////////
region returns [Region r]
@init {
ArrayList<Subregion> subrs = new ArrayList<Subregion>();
}:
^(REGION (subregion {subrs.add($subregion.sr);})+) {$r = new Region(subrs, false);};



const_region returns [Region r]
@init {
ArrayList<Subregion> subrs = new ArrayList<Subregion>();
}: 
^(CONST_REGION (const_subregion {subrs.add($const_subregion.sr);})+) {$r = new Region(subrs, true);};



	
subregion returns [Subregion sr]: 
  const_subregion {$sr = $const_subregion.sr;}
| relative_subregion {$sr = $relative_subregion.sr;};


	
const_subregion returns [Subregion sr]: 
^(CONST_SUBREGION e1=expr e2=expr?) {
if ($e2.text != null) {
  $sr = new Subregion($e1.e,$e2.e,true);
} else {
  $sr = new Subregion($e1.e,$e1.e,true);
}
};



relative_subregion returns [Subregion sr]:
^(RELATIVE_SUBREGION expr) {$sr = new Subregion($expr.e,true);};



////////////////////////////////////////////////////////////////////////////////
// Statements
////////////////////////////////////////////////////////////////////////////////
pf_stmt returns [PointFunctionStatement pfs]: 
local_var_decl 
{
$pfs = new PointFunctionStatement($local_var_decl.lv);
}
| assign_stmt
{
$pfs = new PointFunctionStatement($assign_stmt.as);
};


assign_stmt returns [ExpressionStatement as]:
^(ASSIGN_OP ID expr) 
{
LocalVarRef lhs = new LocalVarRef($ID.text);
Expression rhs = $expr.e;
Assignment a = new Assignment(lhs,rhs);
$as = new ExpressionStatement(a);
}
| ^(ASSIGN_OP gd_ref expr)
{
GridDataRef lhs = $gd_ref.gdr;
Expression rhs = $expr.e;
Assignment a = new Assignment(lhs,rhs);
$as = new ExpressionStatement(a);
}
;
	   


local_var_decl returns [LocalVar lv]:	
  ^(LOCAL_VAR_DECL ID type) 
{
$lv = new LocalVar($ID.text, $type.text);
}
| ^(LOCAL_VAR_DECL ID type expr) 
{
$lv = new LocalVar($ID.text, $type.text, $expr.e);
};



reduce_stmt returns [ReductionStatement rs]
: 
^(REDUCE_STMT ^(SUBGRID region) expr) 
{
$rs = new ReductionStatement($region.r, $expr.e);
};



stencil_stmt returns [StencilStatement ss]: 
^(STENCIL_STMT ^(SUBGRID region) assign_stmt)
{
$ss = new StencilStatement($region.r,$assign_stmt.as);
}
| ^(STENCIL_STMT ^(SUBGRID region) pf_call_stmt)

{
$ss = new StencilStatement($region.r,$pf_call_stmt.pfcs);
};



pf_call_stmt returns [ExpressionStatement pfcs]: pf_call
{
$pfcs = new ExpressionStatement($pf_call.pfc);
};



	
////////////////////////////////////////////////////////////////////////////////
// Expression
////////////////////////////////////////////////////////////////////////////////
expr returns [Expression e]:
      ^(GT_OP  lhs=expr rhs=expr) {$e = new GreaterThan($lhs.e, $rhs.e);}
    | ^(LT_OP  lhs=expr rhs=expr) {$e = new LessThan($lhs.e, $rhs.e);}
    | ^(GTE_OP  lhs=expr rhs=expr) {$e = new GreaterEqThan($lhs.e, $rhs.e);}
    | ^(LTE_OP  lhs=expr rhs=expr) {$e = new LessEqThan($lhs.e, $rhs.e);}
    | ^(EQ_OP  lhs=expr rhs=expr) {$e = new EqualTo($lhs.e, $rhs.e);}
    | ^(NE_OP  lhs=expr rhs=expr) {$e = new NotEqualTo($lhs.e, $rhs.e);}
    | ^(ADD_OP lhs=expr rhs=expr) {$e = new Add($lhs.e, $rhs.e);}
    | ^(SUB_OP lhs=expr rhs=expr) {$e = new Subtract($lhs.e, $rhs.e);}
    | ^(MUL_OP lhs=expr rhs=expr) {$e = new Multiply($lhs.e, $rhs.e);}
    | ^(DIV_OP lhs=expr rhs=expr) {$e = new Divide($lhs.e, $rhs.e);}
    | ^(MOD_OP lhs=expr rhs=expr) {$e = new Mod($lhs.e, $rhs.e);}
    | ^(PWR_OP lhs=expr rhs=expr) {$e = new Power($lhs.e, $rhs.e);}
    | ^(UNARY_PLUS o=expr) {$e = new UnaryPlus($o.e);}
    | ^(UNARY_MINUS o=expr) {$e = new UnaryMinus($o.e);}
    | INT {$e = new IntLiteral(Integer.parseInt($INT.text));}
    | LONG {String scrubbed = $LONG.text; 
            if ($LONG.text.endsWith("l") || $LONG.text.endsWith("L")) {
              int length = $LONG.text.length();
              scrubbed = $LONG.text.substring(0,length - 1);
            }
            $e = new LongLiteral(Long.parseLong(scrubbed));}
    | FLOAT {String scrubbed = $FLOAT.text; 
             if ($FLOAT.text.endsWith("f") || $FLOAT.text.endsWith("F")) {
               int length = $FLOAT.text.length();
               scrubbed = $FLOAT.text.substring(0,length - 1);
             } 
             $e = new FloatLiteral(Float.parseFloat(scrubbed));}
    | DOUBLE {$e = new DoubleLiteral(Double.parseDouble($DOUBLE.text));}
    | ID {$e = new LocalVarRef($ID.text);}
    | pf_call {$e = $pf_call.pfc;}
    | f_call {$e = $f_call.fc;}
    | gd_ref {$e = $gd_ref.gdr;};


	
pf_call returns [PointFunctionCall pfc]
: ^(POINT_FUNCTION_CALL ID ^(ARG_LIST grid_data_list)) {
$pfc = new PointFunctionCall($ID.text, $grid_data_list.gdList);
$pfc.setPointFunction(p.getPointFunction($ID.text));};



f_call returns [FunctionCall fc]
: ^(FUNCTION_CALL ID ^(ARG_LIST arg_list)) {$fc = new FunctionCall($ID.text, $arg_list.eList);};




////////////////////////////////////////////////////////////////////////////////
// References
////////////////////////////////////////////////////////////////////////////////
gd_ref returns [GridDataRef gdr]
@init {
ArrayList<SpaceRef> spaceRefs = new ArrayList<SpaceRef>();
}
: ^(ID ^(TIME_REF time_ref?) ^(SPACE_REF (space_ref {spaceRefs.add($space_ref.sr);})+)) 
{
String name = $ID.text;
GridData gd = new GridData(name);

Expression timeRef = null;
if ($time_ref.text != null) {
  timeRef = $time_ref.e;
}

$gdr = new GridDataRef(gd,timeRef,spaceRefs);
};


time_ref returns [Expression e]: expr {$e = $expr.e;};



space_ref returns [SpaceRef sr]
: const_ref {$sr = $const_ref.sr;}
| relative_ref {$sr = $relative_ref.sr;};	



const_ref returns [SpaceRef sr]
: ^(CONST_REF expr) {$sr = new SpaceRef($expr.e, true);};



relative_ref returns [SpaceRef sr]
: ^(RELATIVE_REF expr {$sr = new SpaceRef($expr.e, false);});

