/*
 * SDSL.g: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */
 
/*
 * @file: SDSL.g
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 */



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// General
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
grammar SDSL;
options {
	backtrack = true;
	memoize = true;
	output = AST;
	ASTLabelType=CommonTree;
}
////////////////////////////////////////////////////////////////////////////////
// AST tokens
////////////////////////////////////////////////////////////////////////////////
tokens {
	SDSL_PROG;
	
	CONST_DEF;
	PARAM_DECL;
	LOCAL_VAR_DECL;
	
	GRID_DEF;
	GRID_DATA_DEF;
	
	POINT_FUNCTION_DEF;
	ITER_DEF;
	STENCIL_DEF;
	REDUCTION_DEF;
	
	REGION;
	CONST_REGION;
	CONST_SUBREGION;
	RELATIVE_SUBREGION;
	
	SUBGRID_COMPUTATION;
	PARAMETER_LIST;
	SUBCALC_LIST;
	
	GRID_DATA_REF;
	TIME_REF;
	SPACE_REF;
	CONST_REF;
	RELATIVE_REF;
	
	POINT_FUNCTION_CALL;
	FUNCTION_CALL;
	EXPR;
	TYPE;
	NAME;
	GRID_DATA_LIST;
	DEFINING_GRID;
	STMT_LIST;
	ASSOC_OP;
	ARG_LIST;
	SUBGRID;
	OPERATION;
	ASSIGN_STMT;
	REDUCE_STMT;
	STENCIL_STMT;
	LHS;
	RHS;
	
	ASSIGN_OP;
	ADD_OP;
	SUB_OP;
	MUL_OP;
	DIV_OP;
	MOD_OP;
	PWR_OP;
	
	GT_OP;
	LT_OP;
	GTE_OP;
	LTE_OP;
	EQ_OP;
	NE_OP;
	
	UNARY_MINUS;
	UNARY_PLUS;
	
	VALUE;
}

@parser::header {
package edu.osu.cse.hpcrl.sdsl.frontend;
}

@lexer::header {
package edu.osu.cse.hpcrl.sdsl.frontend;
}
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Parser rules
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 * Root of a SDSL program
 */
sdsl
: decl_list -> ^(SDSL_PROG decl_list) ;



////////////////////////////////////////////////////////////////////////////////
// Lists
////////////////////////////////////////////////////////////////////////////////
decl_list: decl*;

time_list: expr time_list_elem*;
time_list_elem:	COMMA expr -> expr;

grid_data_list: ID grid_data_list_elem*;
grid_data_list_elem: COMMA ID -> ID;

arg_list: expr arg_list_elem*;
arg_list_elem: COMMA expr -> expr;


iter_list: computable iter_list_elem*;
iter_list_elem: computable;
computable: stencil_def
          | reduce_def;

////////////////////////////////////////////////////////////////////////////////
// Declarations and definitions
////////////////////////////////////////////////////////////////////////////////
decl: param_decl
    | const_def
    | grid_def
    | grid_data_def
    | pf_def 
    | iter_def;


////////////////////////////////////////////////////////////////////////////////
// Parameters / constants
////////////////////////////////////////////////////////////////////////////////
param_decl: type ID SEMI -> ^(PARAM_DECL ID type);
const_def: type ID ASSIGN expr SEMI -> ^(CONST_DEF ID type expr);

type: I_TYPE
    | L_TYPE
    | F_TYPE
    | D_TYPE;



////////////////////////////////////////////////////////////////////////////////
// Grid definition
////////////////////////////////////////////////////////////////////////////////
grid_def: GRID ID const_region SEMI -> ^(GRID_DEF ID const_region);



////////////////////////////////////////////////////////////////////////////////
// Grid Data definition
////////////////////////////////////////////////////////////////////////////////
grid_data_def: type GRID_DATA name=ID ON grid=ID (AT time_list)? SEMI -> ^(GRID_DATA_DEF type $name $grid time_list?);



////////////////////////////////////////////////////////////////////////////////
// Point Function definition
////////////////////////////////////////////////////////////////////////////////
pf_def:	POINT_FUNC ID OPAREN grid_data_list? CPAREN OBRACE pf_stmt+ CBRACE -> ^(POINT_FUNCTION_DEF ID ^(PARAMETER_LIST grid_data_list?) ^(STMT_LIST pf_stmt+)); 



////////////////////////////////////////////////////////////////////////////////
// Iteration definition
////////////////////////////////////////////////////////////////////////////////
iter_def: ITERATE i1=INT OBRACE iter_list CBRACE (CHECK OPAREN expr CPAREN EVERY i2=INT ITERATIONS)? -> ^(ITER_DEF $i1 iter_list expr? $i2?);



////////////////////////////////////////////////////////////////////////////////
// Stencil definition
////////////////////////////////////////////////////////////////////////////////
stencil_def: STENCIL ID OBRACE stencil_stmt+ CBRACE -> ^(STENCIL_DEF ID ^(SUBCALC_LIST stencil_stmt+));



////////////////////////////////////////////////////////////////////////////////
// Reduction definition
////////////////////////////////////////////////////////////////////////////////
reduce_def: REDUCTION ID assoc_op OBRACE reduce_stmt+ CBRACE -> ^(REDUCTION_DEF ID assoc_op ^(SUBCALC_LIST reduce_stmt+));
	
assoc_op: PLUS
        | TIMES
        | MAX
        | MIN;



////////////////////////////////////////////////////////////////////////////////
// Region
////////////////////////////////////////////////////////////////////////////////
region: subregion+ -> ^(REGION subregion+);
const_region: const_subregion+	-> ^(CONST_REGION const_subregion+);
	
subregion: const_subregion|relative_subregion;
	
const_subregion: OBRACK e1=expr (COLON e2=expr)? CBRACK -> ^(CONST_SUBREGION $e1 $e2?);
relative_subregion: OBRACE expr CBRACE -> ^(RELATIVE_SUBREGION expr);



////////////////////////////////////////////////////////////////////////////////
// Statements
////////////////////////////////////////////////////////////////////////////////
pf_stmt: local_var_decl
       | assign_stmt;

stencil_stmt: region COLON assign_stmt -> ^(STENCIL_STMT ^(SUBGRID region) assign_stmt)
            | region COLON pf_call_stmt -> ^(STENCIL_STMT ^(SUBGRID region) pf_call_stmt);

reduce_stmt: region COLON expr SEMI -> ^(REDUCE_STMT ^(SUBGRID region) expr);

assign_stmt: ID ASSIGN expr SEMI -> ^(ASSIGN_OP ID expr)
	         | gd_ref ASSIGN expr SEMI -> ^(ASSIGN_OP gd_ref expr);
	    
pf_call_stmt: pf_call SEMI -> pf_call;

local_var_decl:	type ID SEMI -> ^(LOCAL_VAR_DECL ID type)
	            | type ID ASSIGN expr SEMI -> ^(LOCAL_VAR_DECL ID type expr);

	
	
////////////////////////////////////////////////////////////////////////////////
// Expression
////////////////////////////////////////////////////////////////////////////////
add_op: PLUS -> ADD_OP
      | MINUS -> SUB_OP;
mul_op: TIMES -> MUL_OP
      | DIV -> DIV_OP
      | MOD -> MOD_OP;
pwr_op: PWR -> PWR_OP;

lgt_op: GT-> GT_OP
      | LT -> LT_OP
      | GTE -> GTE_OP
      | LTE -> LTE_OP;
          
enq_op: EQ -> EQ_OP
      | NE -> NE_OP;

expr: enq_expr;
enq_expr: lgt_expr (enq_op^ lgt_expr)*;
lgt_expr: add_expr (lgt_op^ add_expr)*;
add_expr: mul_expr (add_op^ mul_expr)*;
mul_expr: una_expr (mul_op^ una_expr)*;
//pwr_expr: una_expr (pwr_op^ pwr_expr)?;
una_expr: PLUS una_expr -> ^(UNARY_PLUS una_expr)
	| MINUS una_expr -> ^(UNARY_MINUS una_expr)
	| una_expr_not_pm;

una_expr_not_pm: primary;	

primary: OPAREN expr CPAREN -> expr
       | INT
       | LONG
       | FLOAT
       | DOUBLE
       | ID
       | pf_call
       | f_call
       | gd_ref;



pf_call: ID OPAREN grid_data_list CPAREN -> ^(POINT_FUNCTION_CALL ID ^(ARG_LIST grid_data_list));
f_call: ID OPAREN arg_list CPAREN -> ^(FUNCTION_CALL ID ^(ARG_LIST arg_list));
 




////////////////////////////////////////////////////////////////////////////////
// References
////////////////////////////////////////////////////////////////////////////////
gd_ref: (time_ref)? ID space_ref+ -> ^(ID ^(TIME_REF time_ref?) ^(SPACE_REF space_ref+));

time_ref: OBRACK expr CBRACK -> expr;
space_ref: const_ref
         | relative_ref;	


const_ref: OBRACE expr CBRACE -> ^(CONST_REF expr);
relative_ref : OBRACK expr CBRACK -> ^(RELATIVE_REF expr);



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Lexer rules
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


I_TYPE 	: 'int';
L_TYPE  : 'long';
F_TYPE 	: 'float';
D_TYPE  : 'double';


ASSIGN	:	'=';
PLUS	:	'+';
MINUS	:	'-';
TIMES	:	'*';
DIV	:	'/';
MOD	:	'%';
PWR	:	'^';
MAX	:	'max';
MIN	:	'min';

GT: '>';
LT: '<';
GTE: '>=';
LTE: '<=';
EQ: '==';
NE: '!=';

SEMI	:	'\;';
COMMA	:	',';
COLON	:	':';
ON	:	'on';
AT	:	'at';

OBRACK	:	'[';
CBRACK	:	']';
OBRACE	:	'{';
CBRACE	:	'}';
OPAREN	:	'(';
CPAREN	:	')';
//OANGLE	:	'<';
//CANGLE	:	'>';


GRID	:	'grid';
GRID_DATA
	:	'griddata';
POINT_FUNC
	:	'pointfunction';
STENCIL	:	'stencil';
REDUCTION:	'reduction';
ITERATE: 'iterate';
CHECK: 'check';
EVERY: 'every';
ITERATIONS: 'iterations';
	
	
ID  :	('a'..'z'|'A'..'Z'|'_') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*
    ;

INT :	'0'..'9'+
    ;
    
LONG :	'0'..'9'+ 'l'
     ;

FLOAT
    :   ('0'..'9')+ '.' ('0'..'9')* EXPONENT? 'f'
    |   '.' ('0'..'9')+ EXPONENT? 'f'
    |   ('0'..'9')+ EXPONENT 'f'
    ;
    
DOUBLE
    :   ('0'..'9')+ '.' ('0'..'9')* EXPONENT?
    |   '.' ('0'..'9')+ EXPONENT?
    |   ('0'..'9')+ EXPONENT
    ;

COMMENT
    :   '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    |   '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

WS  :   ( ' '
        | '\t'
        | '\r'
        | '\n'
        ) {$channel=HIDDEN;}
    ;

STRING
    :  '"' ( ESC_SEQ | ~('\\'|'"') )* '"'
    ;

fragment
EXPONENT : ('e'|'E') ('+'|'-')? ('0'..'9')+ ;

fragment
HEX_DIGIT : ('0'..'9'|'a'..'f'|'A'..'F') ;

fragment
ESC_SEQ
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   UNICODE_ESC
    |   OCTAL_ESC
    ;

fragment
OCTAL_ESC
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;

fragment
UNICODE_ESC
    :   '\\' 'u' HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT
    ;
