/*
 * SDSLC.java: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */
package edu.osu.cse.hpcrl.sdsl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import edu.osu.cse.hpcrl.sdsl.backend.CodeGenerator;
import edu.osu.cse.hpcrl.sdsl.backend.OvertileCodeGenerator;
import edu.osu.cse.hpcrl.sdsl.frontend.FrontendDriver;
import edu.osu.cse.hpcrl.sdsl.ir.SDSLProgram;
import edu.osu.cse.hpcrl.sdsl.util.BasicFileIO;
import edu.osu.cse.hpcrl.sdsl.util.CommandLineOptions;
import edu.osu.cse.hpcrl.sdsl.util.ExitCodes;
import edu.osu.cse.hpcrl.sdsl.util.OvertileOptions;


/**
 * Main driver class for SDSL source-to-source compiler.
 * 
 * @author Tom Henretty <henretty@cse.ohio-state.edu>
 *
 */
public class SDSLC {
  
  // Logging
  private static Logger log = Logger.getLogger(SDSLC.class.getName());
  
  // Command line options
  private static CommandLineOptions options;
  
  // Input file text
  private static String eSDSLText = "";


  /**
   * Main method drives source-to-source conversion of SDSL to C.
   * 
   * @param args Command line arguments
   */
  public static void main(String[] args) {
    
    // Set up logging
    BasicConfigurator.configure();
    log.setLevel(Level.ERROR);

    // Parse command line arguments
    options = new CommandLineOptions();
    options.parseArgs(args);
    
    // Set log level
    log.setLevel(options.getLogLevel());

    // Parse embedded SDSL sections
    log.debug("Parsing input file...");
    ArrayList<SDSLProgram> sdslList = parseEmbeddedSDSL(options.getInFileName(), options);
    log.debug("...finished parsing input file.");
    
    // Generate calls to optimized C functions
    log.debug("Generating function calls...");
    ArrayList<String> calls = genFunctionCalls(sdslList, options);
    log.debug("...finished generating function calls.");
    
    // Generate optimized C function definitions
    log.debug("Generating function definitions...");
    ArrayList<String> defs = genFunctionDefs(sdslList, options);
    log.debug("...finished generating function definitions.");
    
    // Write the optimized eSDSL file
    log.debug("Writing optimized file...");
    writeOptimizedC(calls,defs,options);
    log.debug("...finished writing optimized file.");
  }
  

  
  /**
   * Gets all embedded SDSL sections in fileName. Each section is parsed into
   * an SDSLProgram object that contains both full text and IR representations 
   * of the SDSL program.
   *  
   * @param fileName Name of file containing embedded SDSL
   * @return An ArrayList of SDSLProgram objects in the same order they are
   *         encountered in the input file
   */
  private static ArrayList<SDSLProgram> parseEmbeddedSDSL(String fileName, CommandLineOptions options) {
    ArrayList<SDSLProgram> sdslList = new ArrayList<SDSLProgram>();

    //===-----------------------------------------------------------------------
    // Read the input file
    //===-----------------------------------------------------------------------
    try {
      eSDSLText = BasicFileIO.readFile(fileName);
    } catch (IOException e) {
      e.printStackTrace();
      log.fatal(fileName + ": COULD NOT READ");
      System.exit(ExitCodes.FILE_READ_FAILED);
    }

    //===-----------------------------------------------------------------------
    // Pull SDSL sections from the file via regex
    //===-----------------------------------------------------------------------
    int numSDSL = 0;
    try {
      String sdslRegex = "(#pragma\\s+sdsl\\s+begin.*?#pragma\\s+sdsl\\s+end)";
      Pattern sdslPattern = Pattern.compile(sdslRegex, Pattern.DOTALL);
      Matcher sdslMatcher = sdslPattern.matcher(eSDSLText);

      while (sdslMatcher.find()) {
        String sdsl = sdslMatcher.group();
        SDSLProgram p = parseSDSLSection(sdsl, options);
        p.setSDSLIndex(numSDSL);
        sdslList.add(p);
        numSDSL++;
      }
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("SDSL EXTRACTION FAILED (SECTION " + numSDSL + ")");
      System.exit(ExitCodes.PARSE_FAILED);
    }
    
    return sdslList;
  }
  
  /**
   * 
   * @param sdsl
   * @return
   */
  private static SDSLProgram parseSDSLSection(String sdsl, CommandLineOptions options) throws Exception {
    String beginPragma = "";

    StringBuffer noStartSDSL = new StringBuffer();
    String cleanSDSL = "";
    
    //===-----------------------------------------------------------------------
    // Strip begin and end pragmas
    //
    // If we're generating overlapped tiling code grab the options from the 
    // begin pragma
    //===-----------------------------------------------------------------------
    
    // FIXME: This should not require 2 passes over the SDSL
    Pattern beginPattern = Pattern.compile("^(#pragma\\s+sdsl\\s+begin.*)$",Pattern.MULTILINE);
    Matcher beginMatcher = beginPattern.matcher(sdsl.trim());
    while (beginMatcher.find()) {
      // FIXME: Bail if more than one start pragma found
      beginPragma = beginMatcher.group();
      beginMatcher.appendReplacement(noStartSDSL,"");
    }
    beginMatcher.appendTail(noStartSDSL);
    OvertileOptions otOptions = null;
    if (options.getCodeGenerator() instanceof OvertileCodeGenerator) {
      otOptions = new OvertileOptions(beginPragma);
    }
    
    Pattern endPattern = Pattern.compile("(#pragma\\s+sdsl\\s+end.*)",Pattern.MULTILINE);
    Matcher endMatcher = endPattern.matcher(noStartSDSL);
    cleanSDSL = endMatcher.replaceAll("");

    //===-----------------------------------------------------------------------
    // Parse
    //===-----------------------------------------------------------------------
    SDSLProgram p = FrontendDriver.parseToSDSLProgram(cleanSDSL,options);
    p.setSDSLText(sdsl);
    p.setCLOptions(options);
    p.setOTOptions(otOptions);
    return p;
  }
  
  private static ArrayList<String> genFunctionCalls(ArrayList<SDSLProgram> sdslList, CommandLineOptions options) {
    ArrayList<String> calls = new ArrayList<String>();
    CodeGenerator codeGen = options.getCodeGenerator();
    for (SDSLProgram p : sdslList) {
      String call = "";
      call += codeGen.generateCFunctionPrototype(p) + ";\n";
      call += codeGen.generateCFunctionCall(p);
      calls.add(call);
    }
    return calls;
  }
  
  private static ArrayList<String> genFunctionDefs(ArrayList<SDSLProgram> sdslList, CommandLineOptions options) {
    ArrayList<String> defs = new ArrayList<String>();
    CodeGenerator codeGen = options.getCodeGenerator();
    for (SDSLProgram p : sdslList) {
      String def = codeGen.generateCFunctionDefinition(p);
      defs.add(def);
    }
    return defs;
  }
  
  private static void writeOptimizedC(ArrayList<String> calls, ArrayList<String> defs, CommandLineOptions options) {
    StringBuffer out = new StringBuffer();
    int numSDSL = 0;
    try {
      String sdslRegex = "(#pragma\\s+sdsl\\s+begin.*?#pragma\\s+sdsl\\s+end)";
      Pattern sdslPattern = Pattern.compile(sdslRegex, Pattern.DOTALL);
      Matcher sdslMatcher = sdslPattern.matcher(eSDSLText);

      while (sdslMatcher.find()) {
        sdslMatcher.appendReplacement(out, calls.get(numSDSL));
        numSDSL++;
      }
      sdslMatcher.appendTail(out);
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("SDSL REPLACEMENT FAILED (SECTION " + numSDSL + ")");
      System.exit(ExitCodes.SDSL_REPLACEMENT_FAILED);
    }
    
    out.append(options.getCodeGenerator().generateCPreprocessorInfo());
    for (String def : defs) {
      out.append(def);
    }
    try {
      BasicFileIO.writeFile(options.getOutFileName(), out.toString());
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("WRITE TO " + options.getOutFileName() + " FAILED");
      System.exit(ExitCodes.FILE_WRITE_FAILED);
    }  
  }
}
