/*
 * OvertileCodeGenerator.java: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */
package edu.osu.cse.hpcrl.sdsl.backend;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;

import org.apache.log4j.Logger;

import edu.osu.cse.hpcrl.sdsl.ir.Add;
import edu.osu.cse.hpcrl.sdsl.ir.Assignment;
import edu.osu.cse.hpcrl.sdsl.ir.BinaryOp;
import edu.osu.cse.hpcrl.sdsl.ir.Computable;
import edu.osu.cse.hpcrl.sdsl.ir.Constant;
import edu.osu.cse.hpcrl.sdsl.ir.Divide;
import edu.osu.cse.hpcrl.sdsl.ir.DoubleLiteral;
import edu.osu.cse.hpcrl.sdsl.ir.EqualTo;
import edu.osu.cse.hpcrl.sdsl.ir.Expression;
import edu.osu.cse.hpcrl.sdsl.ir.ExpressionStatement;
import edu.osu.cse.hpcrl.sdsl.ir.FloatLiteral;
import edu.osu.cse.hpcrl.sdsl.ir.FunctionCall;
import edu.osu.cse.hpcrl.sdsl.ir.GreaterEqThan;
import edu.osu.cse.hpcrl.sdsl.ir.GreaterThan;
import edu.osu.cse.hpcrl.sdsl.ir.Grid;
import edu.osu.cse.hpcrl.sdsl.ir.GridData;
import edu.osu.cse.hpcrl.sdsl.ir.GridDataRef;
import edu.osu.cse.hpcrl.sdsl.ir.IntLiteral;
import edu.osu.cse.hpcrl.sdsl.ir.Iteration;
import edu.osu.cse.hpcrl.sdsl.ir.LessEqThan;
import edu.osu.cse.hpcrl.sdsl.ir.LessThan;
import edu.osu.cse.hpcrl.sdsl.ir.LocalVar;
import edu.osu.cse.hpcrl.sdsl.ir.LocalVarRef;
import edu.osu.cse.hpcrl.sdsl.ir.LongLiteral;
import edu.osu.cse.hpcrl.sdsl.ir.Multiply;
import edu.osu.cse.hpcrl.sdsl.ir.NotEqualTo;
import edu.osu.cse.hpcrl.sdsl.ir.Parameter;
import edu.osu.cse.hpcrl.sdsl.ir.PointFunction;
import edu.osu.cse.hpcrl.sdsl.ir.PointFunctionCall;
import edu.osu.cse.hpcrl.sdsl.ir.PointFunctionStatement;
import edu.osu.cse.hpcrl.sdsl.ir.Reduction;
import edu.osu.cse.hpcrl.sdsl.ir.ReductionStatement;
import edu.osu.cse.hpcrl.sdsl.ir.Region;
import edu.osu.cse.hpcrl.sdsl.ir.SDSLProgram;
import edu.osu.cse.hpcrl.sdsl.ir.SpaceRef;
import edu.osu.cse.hpcrl.sdsl.ir.Statement;
import edu.osu.cse.hpcrl.sdsl.ir.Stencil;
import edu.osu.cse.hpcrl.sdsl.ir.StencilStatement;
import edu.osu.cse.hpcrl.sdsl.ir.Subregion;
import edu.osu.cse.hpcrl.sdsl.ir.Subtract;
import edu.osu.cse.hpcrl.sdsl.ir.UnaryMinus;

// FIXME: This file is an atrocious hack of code Justin and Tom wrote.
// It really needs to be cleaned up and functionality shared with the 
// affine code generator should be moved to the superclass. Also,
// documentation is horrid and should be improved.

// FIXME: Check that all situations where int/long/float/doubles are output to 
// SSP can actually be parsed by SSP parser
/**
 * @author Tom Henretty <henretty@cse.ohio-state.edu>
 * @author Justin Holewinski <justin.holewinski@gmail.com>
 */
public class OvertileCodeGenerator implements CodeGenerator {
  private static final String EOL = System.getProperty("line.separator");
  private static final String TAB = "  ";
  
  private static final String[] spaceIters = {"i","j","k"};
  private static       Logger   log        = Logger.getLogger(OvertileCodeGenerator.class.getName());

  private StringBuffer sspOut         = new StringBuffer();
  private SDSLProgram  Prog;
  private StringBuffer GlobalLetBuffer = new StringBuffer();
  private String       CurrentField    = "";
  private String       Source          = "<unknown>";
  private String       ssp             = "";
  private String       ProgramName     = "sdsl_program";
  private int          maxDim          = -1;
  
  public OvertileCodeGenerator() {
    super();
  }
  
  public String generateCFunctionPrototype(SDSLProgram p) {
    StringBuffer out = new StringBuffer();
    Source = p.getSDSLText();
    out.append(indent(1) + "int " + genFunctionName() + "(");
    //maxDim = 0;
    
    // Parameter for each grid data
    boolean firstParam = true;
    for (GridData gd : p.getGridData()) {
      int dims = gd.getDefiningGrid().getNumDims();
      maxDim = maxDim > dims ? maxDim : dims;
      String comma = firstParam ? "" : ", ";
      out.append(comma + gd.getType() + " *p_" + gd.getName());
      firstParam = false;
    }
    
    // Parameter for each SDSL parameter
    for (Parameter param : p.getParams()) {
      out.append(", " + param.getType() + " sdsl_" + param.getName());
    }

    out.append(")");
    return out.toString();
  }

  public String generateCFunctionCall(SDSLProgram p) {
    StringBuffer out = new StringBuffer();
    Source = p.getSDSLText();
    out.append(indent(1) + "const int sdsl_return_" + p.getSDSLIndex() + " = " + genFunctionName() + "(");
    
    // Parameter for each grid data
    boolean firstParam = true;
    for (GridData gd : p.getGridData()) {
      String comma = firstParam ? "" : ", ";
      out.append(comma + gd.getName());
      firstParam = false;
    }
    
    // Parameter for each SDSL parameter
    for (Parameter param : p.getParams()) {
      out.append(", "  + param.getName());
    }
    out.append(");\n");
    return out.toString();
  }

  public String generateCFunctionDefinition(SDSLProgram p) {
    StringBuffer out = new StringBuffer();
    
    //--------------------------------------------------------------------------
    // Overtile generates device, then host function
    //--------------------------------------------------------------------------
    Source = p.getSDSLText();
    sspOut = new StringBuffer();
    
    ArrayList<Integer> spaceTileSizes = p.getOTOptions().getSpaceTileSizes();
    ArrayList<Integer> threadBlockSizes = p.getOTOptions().getThreadBlockSizes();
    int time = p.getOTOptions().getTimeTileSize();
    boolean verbose = p.getCLOptions().outputVerbose();
    boolean legacy = p.getCLOptions().isLegacyGPU();

    // Convert space tile and thread block size array lists to primitive arrays 
    // for easier JNI access 
    int stSizesArr[] = new int[spaceTileSizes.size()];
    int i = 0;
    for (Integer stSize : spaceTileSizes) {
      stSizesArr[i] = stSize.intValue();
      i++;
    }
    
    int tbSizesArr[] = new int[threadBlockSizes.size()];
    i = 0;
    for (Integer tbSize : threadBlockSizes) {
      tbSizesArr[i] = tbSize.intValue();
      i++;
    }
    ssp = genCode(p);
    if (verbose) {
      System.err.print("*** SSP START ***\n" + ssp + "*** SSP END ***\n");
    }
    out.append(LibOvertileWrapper.generateCUDAFunctionDefinition(ssp, tbSizesArr, stSizesArr, time, legacy, verbose));
    
    //--------------------------------------------------------------------------
    // We generate a c function to wrap the overtile-generated host function.
    // This C function contains outer time loops and reductions
    //--------------------------------------------------------------------------
    long maxIter = p.getMaxIterations();
    
    // checkEvery == Time tile size, otSize == Overtile time tile size
    // If checkEvery % otSize != 0 there may be correctness issues
    int checkEvery = (int)p.getIteration().getCheckEvery();
    int otSize = p.getOTOptions().getTimeTileSize();
    if (checkEvery == 0) checkEvery = otSize;
    
    // Do we have converge checks or reductions?
    boolean hasCheck = p.getIteration().getConvergeCheck() != null;
    boolean hasReduction = false;
    for (Computable c: p.getIteration().getComputables()) {
      if (c instanceof Reduction) {
        hasReduction = true;
        break;
      }
    }
    
    
    // Start wrapper function def
    if (p.getCLOptions().isExternC()) {
      out.append("#ifdef __cplusplus\nextern \"C\"\n#endif\n");
    }
    out.append(generateCFunctionPrototype(p) + " {" + EOL);
    out.append(indent(1) + "int tt,t,i,j,k,ret = 0;" + EOL);
    out.append(genArrayDecls());
    out.append(genCasts());
    out.append(genConstants());
    out.append(genReductionVars());
    

    // Two cases:
    // 1) transfer optimized
    // 2) transfer semi-optimized
    
    if (!(hasCheck || hasReduction)) {
      // Case 1) Time tile loop in overtile generated code
      out.append(genOTFunctionCall(p, (int)maxIter));
    } else if ((hasCheck || hasReduction)) {
      // Case 2) Time tile loops in SDSL and overtile generated codes
      out.append(indent(1) + "for (tt = 0; tt < " + maxIter + "; tt += " + checkEvery + ") {" + EOL);
      out.append(indent(2) + "for (t = tt; t < min(tt+" + checkEvery + "," + maxIter + "); t += " + checkEvery + ") {" + EOL);
      out.append(genArrayCopies()); // FIXME: This could be replaced by a pointer swap to speed things up
      out.append(genOTFunctionCall(p, (int)checkEvery));
      out.append(indent(2) + "} // End time loop" + EOL);
      ArrayList<Computable> Comps = p.getIteration().getComputables();
      for (Computable C : Comps) {
        if (C instanceof Reduction) {
          out.append(genReduction((Reduction)C));
        } 
      }
      if (hasCheck) {
        out.append(indent(3) + genCheck());
      } 
      out.append(indent(1) + "} // End time tile loop" + EOL);
    }
    out.append(indent(0) + "return ret;" + EOL + "} // End function definition" + EOL);
    return out.toString();
  }

  private String genOTFunctionCall(SDSLProgram p, int timesteps) {
    StringBuffer out = new StringBuffer();

    // Host function call
    out.append(indent(3) + "ret = ot_program_" + ProgramName + "(" + timesteps);
    
    // Grid data
    for (GridData gd : p.getGridData()) {
      out.append(", p_" + gd.getName());
    }
    
    // Parameters
    for (Parameter param : p.getParams()) {
      out.append(", sdsl_" + param.getName());
    }
    out.append(");\n");
    return out.toString();
  }
  public String generateCPreprocessorInfo() {
    StringBuffer out = new StringBuffer();
    out.append("#include <float.h>" + EOL);
    return out.toString();
  }
  
  // http://stackoverflow.com/questions/4895523/java-string-to-sha1
  private static String byteArrayToHexString(byte[] b) {
    String result  = "";
    for (int i = 0; i < b.length; i++) {
          result  += Integer.toString((b[i] & 0xff) + 0x100, 16).substring(1);
    }
    return result;
  }
  
  private String genCode(SDSLProgram P) {

    Prog = P;

    MessageDigest MD          = null;
    
    try {
      MD          = MessageDigest.getInstance("SHA1");
      ProgramName = "sdsl_program_" + byteArrayToHexString(MD.digest(Source.getBytes()));
    } catch (NoSuchAlgorithmException E) {
    }
    
    // Header
    sspOut.append("program " + ProgramName + " is\n");

    // Grid
    ArrayList<Grid> Grids = Prog.getGrids();
    if (Grids.size() != 1) {
      log.fatal("Only one grid definition can be present");
      System.exit(1);
    }

    sspOut.append("grid " + Grids.get(0).getNumDims() + "\n");

    // Parameters
    for (Parameter Param : Prog.getParams()) {
      // HACK!
      // FIXME: Overtile code generator doesn't support integer or long parameters
      if (!(Param.getType().equals("int") || Param.getType().equals("long")) ) {
        sspOut.append("param " + Param.getName() + " " + Param.getType() + "\n");
      }
    }
    
    // GridData
    ArrayList<GridData> GridDatas = Prog.getGridData();
    for (GridData GD : GridDatas) {
      // @FIXME: Only inout is derivable from TNPL at the moment
      sspOut.append("field " + GD.getName() + " " + GD.getType() + " inout\n");
    }

    // Constants
    for (Constant C : Prog.getConsts()) {
      StringBuffer OldBuffer = sspOut;
      sspOut = GlobalLetBuffer;
      sspOut.append("let " + C.getName() + " = ");
      genNakedExpression(C.getInitExpr());
      sspOut.append(" in\n");
      sspOut = OldBuffer;
    }

    genIteration(Prog.getIteration());
    
    return sspOut.toString();
  }


  private void genIteration(Iteration Iter) {
    ArrayList<Computable> Comps = Iter.getComputables();
    for (Computable C : Comps) {
      if (C instanceof Stencil) {
        genStencil((Stencil)C);
      } 
    }
  }
  /**
   * Casts pointers to multidimensional array references
   */
  private String genCasts() {
    StringBuffer out = new StringBuffer();
    
    for (GridData gd : Prog.getGridData()) {
      // Get the defining grid and size information
      Grid grid = gd.getDefiningGrid();
      int numDims = grid.getNumDims();
      ArrayList<Subregion> dimSizes = grid.getBounds().getSubregions();
      
      // Build multidim array declaration
      String cast = TAB + gd.getType() + " (*" + gd.getName() + ")";
      for (int dim = 1; dim < numDims; dim++) {
        cast += "[" + exprToString(dimSizes.get(dim).getUB()) + "]";
      }
      
      // Build parameter cast
      cast += " = (" + gd.getType() + " (*)";
      for (int dim = 1; dim < numDims; dim++) {
        cast += "[" + exprToString(dimSizes.get(dim).getUB()) + "]";
      }
      cast += ")p_" + gd.getName() + ";" + EOL;
      out.append(cast);
    }
    return out.toString();
  }
  
  /**
   * Generate array declarations for grid data defined on multiple timesteps 
   */
  private String genArrayDecls() {
    StringBuffer out = new StringBuffer();
    String decls = "";
    for (GridData gd : Prog.getGridData()) {
      // Get the defining grid and size information
      Grid grid = gd.getDefiningGrid();
      int numDims = grid.getNumDims();
      ArrayList<Subregion> dimSizes = grid.getBounds().getSubregions();
      int timeDimSize = gd.getDefinedTimes().size();
      
      if (timeDimSize == 2) {
        // Generate multidim array declaration
        decls += TAB + gd.getType() + " (*" + gd.getName() + "_old)";
        for (int dim = 1; dim < numDims; dim++) {
          decls += "[" + exprToString(dimSizes.get(dim).getUB()) + "]";
        }
        
        // Generate multidim array definition
        decls += " = " + "(" + gd.getType() + " (*)";
        for (int dim = 1; dim < numDims; dim++) {
          decls += "[" + exprToString(dimSizes.get(dim).getUB()) + "]";
        }
        
        // Generate malloc
        decls += ")malloc(";
        for (int dim = 0; dim < numDims; dim++) {
          decls += exprToString(dimSizes.get(dim).getUB()) + "*";
        }
        decls += "sizeof(" + gd.getType() + "));" + EOL;
      }
    }
    out.append(decls);
    return out.toString();
  }
  
  /**
   * Copy original grid data contents into all arrays representing that grid 
   * data. 
   */
  private String genArrayCopies() {
    StringBuffer out = new StringBuffer();
    out.append(EOL + indent(2) + "// [SDSL] Begin array copies" + EOL);
    String copies = "";
    int depth;
    for (depth = 0; depth < maxDim; depth++) {
      String iter = spaceIters[depth];
      Subregion sr = Prog.getGrids().get(0).getBounds().getSubregions().get(depth);
      copies += indent(depth+2) + "for (" + iter + " = " + 0 + "; " +
                                    iter + " < " + exprToString(sr.getUB()) + "; " +iter + "++) {" + EOL;
    }
    
    for (GridData gd : Prog.getGridData()) {
      int timeDimSize = gd.getDefinedTimes().size();
      if (timeDimSize == 2) {
        copies += indent(depth+2) + gd.getName() + "_old";
        for (int dim = 0; dim < gd.getDefiningGrid().getNumDims(); dim++) {
          copies += "[" + spaceIters[dim] + "]";
        }
        copies += " = " + gd.getName();
        for (int dim = 0; dim < gd.getDefiningGrid().getNumDims(); dim++) {
          copies += "[" + spaceIters[dim] + "]";
        }
        copies += ";" + EOL;
      }
    }
    
    
    for (; depth > 0; depth--) {
      copies += indent(depth+1) + "}" + EOL;
    }
    out.append(copies);
    out.append(indent(2) + "// [SDSL] End array copies" + EOL + EOL);
    return out.toString();
  }
  
  private String genReduction(Reduction r) {
    StringBuffer out = new StringBuffer();
    int depth = 0;

    String assocOp = r.getAssocOp();
    String name = r.getName();

    ArrayList<ReductionStatement> rsList = r.getStatements();
    out.append(indent(depth+2) + "// [SDSL] Begin reduction" + EOL);
    // FIXME: Hardcoded reduction var type to double. There needs to be a way
    // to determine what this type should be.
    out.append(indent(depth+2) + "sdsl_" + name + " = ");
    if (assocOp.equals("max")) {
      out.append("-DBL_MAX;" + EOL);
    } else if (assocOp.equals("min")) {
      out.append("DBL_MAX;" + EOL);
    } else if (assocOp.equals("+")) {
      out.append("0.0;" + EOL);
    } else if (assocOp.equals("*")) {
      out.append("1.0;" + EOL);
    }
    for (ReductionStatement rs : rsList) {
      ArrayList<Subregion> subregions = rs.getRegion().getSubregions();

      // Loop nests
      for (Subregion sr : subregions) {
        String iter = spaceIters[depth];
        out.append(indent(depth+2) + "for (" + iter + " = " + exprToString(sr.getLB()) + "; " +
            iter + " <= " + exprToString(sr.getUB()) + "; " + iter + "++) {" + EOL);
        depth++;
      }

      // Statements
      if (assocOp.equals("+") || assocOp.equals("*")) {
        out.append(indent(depth+2) + "sdsl_" + name + " = " + "sdsl_" + name + " " + assocOp + exprToString(rs.getExpression()) + ";");
      }
      else if (assocOp.equals("max") || assocOp.equals("min")) {
        out.append(indent(depth+2) + "sdsl_" + name + " = " + assocOp + "(" + "sdsl_" + name + ", " + exprToString(rs.getExpression()) + ");");
      }

      // Close loop nests
      for (Subregion sr : subregions) {
        depth--;
        out.append(EOL + indent(depth+2) + "}");
      }
      out.append(EOL);
    }
    out.append(indent(depth+2) + "// [SDSL] End reduction" + EOL);
    return out.toString();
  }

  private String genCheck() {
    StringBuffer out = new StringBuffer();
    Expression check = Prog.getIteration().getConvergeCheck();
    if (check != null) {
      out.append("if (" + exprToString(check) + ") break;" + EOL);
    }
    return out.toString();
  }

  /**
   * Generate C language versions of SDSL constants.
   */
  private String genConstants() {
    StringBuffer out = new StringBuffer();
    for (Constant cst : Prog.getConsts()) {
      out.append(TAB + "const " + cst.getType() + " sdsl_" + cst.getName() + 
                 " = " + exprToString(cst.getInitExpr()) + ";" + EOL);
    }
    return out.toString();
  }
  
  /**
   * Generate C language versions of reduction variables.
   */
  private String genReductionVars() {
    StringBuffer out = new StringBuffer();
    for (Computable c : Prog.getIteration().getComputables()) {
      if (c instanceof Reduction) {
        Reduction r = (Reduction)c;
        String assocOp = r.getAssocOp();
        // FIXME: Need to determine type based on expressions in reduction
        String type = "double";
        out.append(type + " sdsl_" + r.getName() + " = ");
        if (assocOp.equals("max")) {
          out.append("-DBL_MAX;" + EOL);
        } else if (assocOp.equals("min")) {
          out.append("DBL_MAX;" + EOL);
        } else if (assocOp.equals("+")) {
          out.append("0.0;" + EOL);
        } else if (assocOp.equals("*")) {
          out.append("1.0;" + EOL);
        }
      }
    }
    return out.toString();
  }
  
  private void genStencil(Stencil Sten) {
    ArrayList<StencilStatement> Stmts = Sten.getStatements();
    for (StencilStatement S : Stmts) {

      Statement Stmt = S.getStatement();
      if (!(Stmt instanceof ExpressionStatement)) {
        log.fatal("Stencil statement must be expression statement");
        System.exit(1);
      }
      Expression Expr = ((ExpressionStatement)Stmt).getExpression();
      if (Expr instanceof Assignment) {
        Assignment A = (Assignment)Expr;
        Expression Assignee = A.getLHS();
        Expression ValueExpr = A.getRHS();
        if (!(Assignee instanceof GridDataRef)) {
          log.fatal("LHS of stencil assignment must be grid data ref");
          System.exit(1);
        }

        GridDataRef GDR = (GridDataRef)Assignee;

        if (!CurrentField.equals(GDR.getGridData().getName())) {
          
          CurrentField = GDR.getGridData().getName();
          sspOut.append(CurrentField + " =\n");
        }

        genBoundRegion(S.getRegion());

        sspOut.append(" : ");

        genNakedExpression(ValueExpr);
        sspOut.append("\n\n");
      } else if (Expr instanceof PointFunctionCall) {
        PointFunctionCall PFC = (PointFunctionCall)Expr;
        PointFunction PF = PFC.getPointFunction();
        HashMap<String, GridData> Bindings = new HashMap<String, GridData>();
        for (int i = 0, e = PF.getParams().size(); i != e; ++i) {
          GridData Param = PF.getParams().get(i);
          Bindings.put(Param.getName(), PFC.getArgs().get(i));
        }
        genPointFunction(PF, Bindings, S.getRegion());
        sspOut.append("\n\n");
      } else {
        log.fatal("Stencil statement must be assignment or point function call");
        System.exit(1);
      }
    }
  }

  private void genPointFunction(PointFunction PF,
                                     HashMap<String, GridData> Bindings,
                                     Region R) {

    StringBuffer LetBuffer = new StringBuffer();
    
    for (PointFunctionStatement PFS : PF.getStmts()) {
      Statement Stmt = PFS.getStatement();

      if (Stmt instanceof LocalVar) {
        // Turn this into a let binding
        LocalVar LV = (LocalVar)Stmt;
        if (LV.getInitExpr() != null) {
          StringBuffer OldBuffer = sspOut;
          sspOut = LetBuffer;
          sspOut.append("let " + LV.getName() + " = ");
          genExpression(LV.getInitExpr(), Bindings);
          sspOut.append(" in\n");
          sspOut = OldBuffer;
        }
      } else if (Stmt instanceof ExpressionStatement) {
        ExpressionStatement ExprStmt = (ExpressionStatement)Stmt;
        if (!(ExprStmt.getExpression() instanceof Assignment)) {
          log.fatal("Point function statement must be assignment");
          System.exit(1);
        }
        Assignment A = (Assignment)(ExprStmt.getExpression());

        if (A.getLHS() instanceof LocalVarRef) {
          // Turn this into a let binding
          StringBuffer OldBuffer = sspOut;
          sspOut = LetBuffer;
          LocalVarRef LVR = (LocalVarRef)(A.getLHS());
          sspOut.append("let " + LVR.getName() + " = ");
          genExpression(A.getRHS(), Bindings);
          sspOut.append(" in\n");
          sspOut = OldBuffer;
        } else if (A.getLHS() instanceof GridDataRef) {
          GridDataRef GDR = (GridDataRef)(A.getLHS());
          String Name = GDR.getGridData().getName();
          if (Bindings.containsKey(Name)) {
            Name = Bindings.get(Name).getName();
          }

          if (!CurrentField.equals(Name)) {
          
            CurrentField = Name;
            sspOut.append(Name + " =\n");
          }

          genBoundRegion(R);

          sspOut.append(" : ");

          sspOut.append(GlobalLetBuffer.toString());
          sspOut.append(LetBuffer.toString());
          
          genExpression(A.getRHS(), Bindings);
          sspOut.append("\n");          
        } else {
          log.fatal("Unhandled LHS in point function statement: " + A.getLHS().getClass().toString());
          System.exit(1);
        }
      }
    }
  }
                                     
  private void genNakedExpression(Expression Expr) {
    genExpression(Expr, new HashMap<String, GridData>());
  }
  
  private void genExpression(Expression Expr, HashMap<String, GridData> Bindings) {
    if (Expr instanceof GridDataRef) {
      GridDataRef GDR = (GridDataRef)Expr;
      String Name = GDR.getGridData().getName();
      if (Bindings.containsKey(Name)) {
        Name = Bindings.get(Name).getName();
      }

      sspOut.append(Name);

      ArrayList<SpaceRef> SpaceRefs = GDR.getSpaceRefs();
      for (SpaceRef SR : SpaceRefs) {
        Expression SRExpr = SR.getRef();
        if (SRExpr instanceof IntLiteral) {
          IntLiteral RefValue = (IntLiteral)SRExpr;
          sspOut.append("[" + RefValue.getValue() + "]");
        } else if (SRExpr instanceof LongLiteral) {
          LongLiteral RefValue = (LongLiteral)SRExpr;
          sspOut.append("[" + RefValue.getValue() + "]");
        } else if (SRExpr instanceof UnaryMinus) {
          UnaryMinus UM = (UnaryMinus)SRExpr;
          Expression Op = UM.getOperand();
          if (Op instanceof IntLiteral) {
            sspOut.append("[-" + ((IntLiteral)Op).getValue() + "]");
          } else if (Op instanceof LongLiteral) {
            sspOut.append("[-" + ((LongLiteral)Op).getValue() + "]");
          } else {
            log.fatal("Space ref must be int literal: " + Op.toString());
            System.exit(1);
          }
        } else {
          log.fatal("Space ref must be int literal: " + SRExpr.toString());
          System.exit(1);
        }
      }
    } else if (Expr instanceof IntLiteral) {
      sspOut.append(((IntLiteral)Expr).getValue());
    } else if (Expr instanceof LongLiteral) {
      sspOut.append(((LongLiteral)Expr).getValue());
    } else if (Expr instanceof FloatLiteral) {
      sspOut.append(((FloatLiteral)Expr).getValue());
    } else if (Expr instanceof DoubleLiteral) {
      sspOut.append(((DoubleLiteral)Expr).getValue());
    } else if (Expr instanceof Add) {
      BinaryOp E = (BinaryOp)Expr;
      sspOut.append("(");
      genExpression(E.getLHS(), Bindings);
      sspOut.append("+");
      genExpression(E.getRHS(), Bindings);
      sspOut.append(")");
    } else if (Expr instanceof Subtract) {
      BinaryOp E = (BinaryOp)Expr;
      sspOut.append("(");
      genExpression(E.getLHS(), Bindings);
      sspOut.append("-");
      genExpression(E.getRHS(), Bindings);
      sspOut.append(")");
    } else if (Expr instanceof Multiply) {
      BinaryOp E = (BinaryOp)Expr;
      sspOut.append("(");
      genExpression(E.getLHS(), Bindings);
      sspOut.append("*");
      genExpression(E.getRHS(), Bindings);
      sspOut.append(")");
    } else if (Expr instanceof Divide) {
      BinaryOp E = (BinaryOp)Expr;
      sspOut.append("(");
      genExpression(E.getLHS(), Bindings);
      sspOut.append("/");
      genExpression(E.getRHS(), Bindings);
      sspOut.append(")");
    } else if (Expr instanceof LocalVarRef) {
      LocalVarRef LVR = (LocalVarRef)Expr;
      sspOut.append(LVR.getName());
    } else if (Expr instanceof FunctionCall) {
      FunctionCall FC = (FunctionCall)Expr;
      sspOut.append(FC.getName() + "(");
      ArrayList<Expression> Params = FC.getArgs();
      for (int i = 0, e = Params.size(); i != e; ++i) {
        if (i > 0)
          sspOut.append(", ");
        genExpression(Params.get(i), Bindings);
      }
      sspOut.append(")");
    } else {
      log.fatal("Unhandled expression type: " + Expr.getClass().toString());
      System.exit(1);
    }
  }
  
  private void genBoundRegion(Region R) {
    sspOut.append("@");
    ArrayList<Subregion> SRs = R.getSubregions();
    for (Subregion SR : SRs) {
      sspOut.append("[");
      genBoundExpr(SR.getLB());
      sspOut.append(":");
      genBoundExpr(SR.getUB());
      sspOut.append("]");
    }
  }
  
  private void genBoundExpr(Expression Expr) {
    if (Expr instanceof IntLiteral) {
      IntLiteral I = (IntLiteral)Expr;
      sspOut.append(I.getValue());
    } else if (Expr instanceof LongLiteral) {
      LongLiteral L = (LongLiteral)Expr;
      sspOut.append(L.getValue());
    } else if (Expr instanceof Subtract) {
      // Must be "var - const"
      Subtract Sub = (Subtract)Expr;
      Expression LHS = Sub.getLHS();
      Expression RHS = Sub.getRHS();
      if (!(LHS instanceof LocalVarRef)) {
        log.fatal("Bound must be 'VAR - CONST'");
        System.exit(1);
      }
      if (!(RHS instanceof IntLiteral || RHS instanceof LongLiteral)) {
        log.fatal("Bound must be 'VAR - CONST'");
        System.exit(1);
      }
      if (RHS instanceof IntLiteral) {
        sspOut.append("$-" + (((IntLiteral)RHS).getValue()-1));
      } else if (RHS instanceof LongLiteral) {
        sspOut.append("$-" + (((LongLiteral)RHS).getValue()-1));
      }
    } else {
      log.fatal("Bound must be constant or subtraction");
      System.exit(1);
    }
  }

  /**
   * Converts an Expression in SDSL IR to a C string representation.
   * @param e
   * @return
   */
  private String exprToString(Expression e) {
    return exprToString(e, new HashMap<String,GridData>());
  }
  
  /**
   * Converts an expression in SDSL IR to a C string representation. Uses 
   * Parameter bindings so that expressions from inside point functions can 
   * be converted depending on the circumstances of their call.
   * 
   * @param e
   * @param bindings
   * @return
   */
  private String exprToString(Expression e, HashMap<String,GridData> bindings) {
    String eStr = "";
    if (e instanceof GridDataRef) {
      GridDataRef gdr = (GridDataRef)e;
      GridData gd = gdr.getGridData();
      String name = gdr.getGridData().getName();
      if (bindings.containsKey(name)) {
        gd = bindings.get(name);
        name = gd.getName();  
      }
      eStr += name;
      
      int gdSize = 0;
      for (GridData gdp : Prog.getGridData()) {
        if (gdp.getName().equals(gd.getName())) {
          gdSize = gdp.getDefinedTimes().size();
          break;
        }
      }
      
      Expression tre = gdr.getTimeRef();
      int tri = 0;
      // FIXME: Check for expressions, long literals
      if (tre instanceof IntLiteral) {
        tri = ((IntLiteral)tre).getValue();
      }
      if (gdSize > 1 && tri == 0) {
        eStr += "_old";
      }
      
      int i = 0;
      for (SpaceRef sr : gdr.getSpaceRefs()) {
        eStr += "[" + spaceIters[i] + "+" +  exprToString(sr.getRef(),bindings) + "]";
        i++;
      }
    } else if (e instanceof IntLiteral) {
      eStr += ((IntLiteral)e).getValue();
    } else if (e instanceof LongLiteral) {
      eStr += ((LongLiteral)e).getValue();
    } else if (e instanceof FloatLiteral) {
      eStr += ((FloatLiteral)e).getValue();
    } else if (e instanceof DoubleLiteral) {
      eStr += ((DoubleLiteral)e).getValue();
    } else if (e instanceof UnaryMinus) {
      eStr += "(-";
      eStr += ((UnaryMinus)e).getOperand();
      eStr += ")";
    } else if (e instanceof Add) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "+";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof Subtract) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "-";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof Multiply) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "*";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof Divide) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "/";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof GreaterThan) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += ">";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof LessThan) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "<";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof GreaterEqThan) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += ">=";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof LessEqThan) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "<=";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof EqualTo) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "==";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof NotEqualTo) {
      BinaryOp o = (BinaryOp)e;
      eStr += "(";
      eStr += exprToString(o.getLHS(), bindings);
      eStr += "!=";
      eStr += exprToString(o.getRHS(), bindings);
      eStr += ")";
    } else if (e instanceof LocalVarRef) {
      LocalVarRef LVR = (LocalVarRef)e;
      eStr += "sdsl_" + LVR.getName();;
    } else if (e instanceof Assignment) {
      BinaryOp o = (BinaryOp)e;
      eStr += exprToString(o.getLHS(), bindings);
      eStr += " = ";
      eStr += exprToString(o.getRHS(), bindings);
    } else if (e instanceof FunctionCall) {
      FunctionCall FC = (FunctionCall)e;
      // HACK
      if (FC.getName().equals("rsqrt")) {
        eStr += "(1.0/sqrt(";
      } else if (FC.getName().equals("rsqrtf")) {
      eStr += "(1.0f/sqrt(";
      }
      else {
        eStr += "(" + FC.getName() + "(";
      }
      // END HACK
      ArrayList<Expression> Params = FC.getArgs();
      for (int i = 0, s = Params.size(); i != s; ++i) {
        if (i > 0)
          eStr += ", ";
        eStr += exprToString(Params.get(i), bindings);
      }
      eStr += "))";
    } else {
      eStr += "/* Unhandled expression */";
    }
    return eStr;
  }
  
  private String indent(int level) {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < level; i++) {
      sb.append(TAB);
    }
    return sb.toString();
  }

  private String genFunctionName() {
    MessageDigest md = null;
    String name = "sdsl_program";
    try {
      md = MessageDigest.getInstance("SHA1");
      name = "sdsl_ot_wrap_" + byteArrayToHexString(md.digest(Source.getBytes()));
    } catch (NoSuchAlgorithmException E) {
    }
    
    return name;
  }
}
