/*
 * FrontendDriver.java: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */
/**
 * 
 */
package edu.osu.cse.hpcrl.sdsl.frontend;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.antlr.runtime.tree.DOTTreeGenerator;
import org.antlr.stringtemplate.StringTemplate;
import org.apache.log4j.Logger;

import edu.osu.cse.hpcrl.sdsl.ir.SDSLProgram;
import edu.osu.cse.hpcrl.sdsl.util.CommandLineOptions;
import edu.osu.cse.hpcrl.sdsl.util.ExitCodes;

/**
 * Drives ANTLR lex, parse to AST, display of AST, and AST-to-SDSLProgram 
 * conversion.
 * 
 * @author Tom Henretty <henretty@cse.ohio-state.edu>
 *
 */
public final class FrontendDriver {
  
  private static Logger log = Logger.getLogger(FrontendDriver.class.getName());
  
  /**
   * Method to orchestrate ANTLR lex, parse to AST, display of AST, and 
   * AST-to-SDSLProgram conversion.
   * 
   * @param sdsl  Text representation of an SDSL program
   * @param options  Command line options
   * @return An SDSLProgram object representing the given SDSL text
   */
  public static SDSLProgram parseToSDSLProgram(String sdsl, CommandLineOptions options) {
    log.setLevel(options.getLogLevel());
    CommonTokenStream tokens = lex(sdsl);
    CommonTree ast = parse(tokens);
    outputAST(ast,options);
    SDSLProgram p = buildIR(ast,tokens);
    return p;
  }
  
  /**
   * Lex the given SDSL text into a token stream.
   * 
   * @param sdsl  Text representation of an SDSL program
   * @return A tokenized version of the given SDSL text
   */
  private static CommonTokenStream lex(String sdsl) {
    CommonTokenStream tokens = null;
    try {
      log.debug("Start lex...");
      log.debug("Creating input stream...");
      ANTLRStringStream input = new ANTLRStringStream(sdsl);
      log.debug("Creating lexer...");
      SDSLLexer lexer = new SDSLLexer(input);
      log.debug("Creating token stream...");
      tokens = new CommonTokenStream(lexer);
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("********* Lexing failed *********");
      System.exit(ExitCodes.LEX_FAILED);
    }
    log.info("...finished lex");
    
    return tokens;
  }
  
  /**
   * Parse the SDSL token stream into an AST.
   * 
   * @param tokens  SDSL token stream
   * @return SDSL AST
   */
  private static CommonTree parse(CommonTokenStream tokens) {
    CommonTree ast = null;
    try {
      log.debug("Start parse to AST...");
      log.debug("Creating parser...");
      SDSLParser parser = new SDSLParser(tokens);
      log.debug("Getting return value...");
      SDSLParser.sdsl_return r = parser.sdsl();
      log.debug("Getting tree...");
      ast = (CommonTree)r.getTree();
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("********* Parsing failed *********");
      System.exit(ExitCodes.PARSE_FAILED);
    }
    log.debug("...finished parse to AST.");
    
    return ast;
  }
  
  /**
   * Outputs the given AST in either plain text or DOT format. Will not output 
   * anything unless it was specified at the command line.
   * 
   * @param ast  The SDSL AST to output
   * @param options  Command line options 
   */
  private static void outputAST(CommonTree ast, CommandLineOptions options) {
    if (options.outputDotAST()) {
      log.debug("Starting to output dot AST... ");
      DOTTreeGenerator gen = new DOTTreeGenerator();
      StringTemplate st = gen.toDOT(ast);
      System.out.println(st);
      log.debug("...finished dot AST output.");
    }
    if (options.outputTextAST()) {
      log.debug("Starting to output text AST... ");
      System.out.println(ast.toStringTree());
      log.debug("...finished text AST output.");
    }
  }
  
  /**
   * Produces the final SDSLProgram object.
   * 
   * @param ast  An SDSL AST
   * @param tokens  An SDSL token stream
   * @return  An SDSLProgram representing the given SDSL AST and token stream
   */
  private static SDSLProgram buildIR(CommonTree ast, CommonTokenStream tokens) {
    SDSLProgram p = null;
    try {
      log.debug("Starting AST->IR convert... ");
      CommonTreeNodeStream nodes = new CommonTreeNodeStream(ast);
      nodes.setTokenStream(tokens);
      ASTParser astParser = new ASTParser(nodes);
      astParser.sdsl();
      p = astParser.getSDSLProgram();
    } catch (Exception e) {
      e.printStackTrace();
      log.fatal("********* AST->IR failed *********");
      System.exit(ExitCodes.AST_IR_CONVERT_FAILED);
    }
    log.debug("...finished AST->IR convert.");
    return p;
  }
}
