/*
 * CommandLineOptions.java: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */
package edu.osu.cse.hpcrl.sdsl.util;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.cli.UnrecognizedOptionException;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import edu.osu.cse.hpcrl.sdsl.backend.AffineCCodeGenerator;
import edu.osu.cse.hpcrl.sdsl.backend.CodeGenerator;
import edu.osu.cse.hpcrl.sdsl.backend.OvertileCodeGenerator;

/**
 * Processes command line arguments using the Apache Commons CLI package.
 * Used as a data object to encapsulate option selections and is passed around 
 * the SDSL compiler.
 * 
 * @author Tom Henretty <henretty@cse.ohio-state.edu>
 *
 */
public final class CommandLineOptions {

  // Logging
  private static Logger log = Logger.getLogger(CommandLineOptions.class.getName());
  
  // Command line options
  private boolean       outputDotAST  = false;
  private boolean       outputTextAST = false;
  private CodeGenerator codeGenerator = null;
  private boolean       outputVerbose = false;
  private String        inFileName    = "";
  private String        outFileName   = "";
  private Level         logLevel      = Level.WARN;
  private boolean       externC       = false;
  private boolean       legacyGPU     = false;
  
  public CommandLineOptions() {
    super();
  }
  
  
  /**
   * Parses command line arguments using Apache Commons CLI.
   * 
   * @param args Command line argument array
   */
  public void parseArgs(String[] args) {
    Options options = new Options();
    options.addOption("h", "help", false, "Print help message and exit");
    //options.addOption("d", "dot-ast", false, "Output AST in dot format");
    //options.addOption("t", "text-ast", false, "Output AST in text format");
    options.addOption("v", "verbose", false, "Output detailed progress");
    options.addOption("o", "out", true, "Output filename");
    options.addOption("f", "in", true, "Embedded SDSL input filename");
    options.addOption("b", "backend", true, "Set backend code generator as 'affine-c' (default) or 'overtile'");
    options.addOption("x", "extern-c", false, "Use C linkage for wrapper functions");
    options.addOption("l", "legacy-gpu", false, "Generate code for NVIDIA GT2xx and Tesla 10xx GPU (overtile backend only)");

    CommandLineParser parser = new PosixParser();
    try {
      CommandLine cmd = parser.parse(options,args);

      if (cmd.hasOption("h")) {
        HelpFormatter hf = new HelpFormatter();
        hf.printHelp("sdslc <options> <input filename>", options);
        System.exit(0);
      }
      
      // Parse back-end
      if (cmd.hasOption("b")) {
        String backend = cmd.getOptionValue('b', "affine-c");
        if (backend.compareTo("affine-c") == 0) {
          codeGenerator = new AffineCCodeGenerator();
        } else if (backend.compareTo("overtile") == 0) {
          codeGenerator = new OvertileCodeGenerator();
        } else {
          System.err.println("Unknown backend: " + backend);
          HelpFormatter hf = new HelpFormatter();
          hf.printHelp("sdslc <options> <input filename>", options);
          System.exit(ExitCodes.BAD_COMMAND_LINE);
        }
      } else {
        // No backend specified, default to affine c
        codeGenerator = new AffineCCodeGenerator();
      }
      
      if (cmd.hasOption("d")) {
        outputDotAST = true;
      }
      if (cmd.hasOption("t")) {
        outputTextAST = true;
      }
      if (cmd.hasOption("v")) {
        outputVerbose = true;
        logLevel = Level.ALL;
      }
      if (cmd.hasOption("f")) {
        inFileName = cmd.getOptionValue("f");
      }
      if (cmd.hasOption("o")) {
        outFileName = cmd.getOptionValue("o");
      } else if (codeGenerator instanceof AffineCCodeGenerator) {
        outFileName = "a.c";
      } else if (codeGenerator instanceof OvertileCodeGenerator) {
        outFileName = "a.cu";
      }
      if (cmd.hasOption("x")) {
        externC = true;
      }
      if (cmd.hasOption("l")) {
        legacyGPU = true;
      }

      
      // Parse positional argument (source file)
      if (!cmd.hasOption("f")) {
        String[] positional = cmd.getArgs();
        if (positional.length == 0 && !cmd.hasOption("f")) {
          System.err.println("No input file specified");
          HelpFormatter hf = new HelpFormatter();
          hf.printHelp("sdslc <options> <input filename>", options);
          System.exit(ExitCodes.BAD_COMMAND_LINE);
        } else if (positional.length > 1) {
          System.err.println("More than one input file specified");
          HelpFormatter hf = new HelpFormatter();
          hf.printHelp("sdslc <options> <input filename>", options);
          System.exit(ExitCodes.BAD_COMMAND_LINE);
        }
        inFileName = positional[0];
      } else if (cmd.hasOption("f") && cmd.getArgs().length > 0) {
        System.err.println("More than one input file specified");
        HelpFormatter hf = new HelpFormatter();
        hf.printHelp("sdslc <options> <input filename>", options);
        System.exit(ExitCodes.BAD_COMMAND_LINE);
      }
      
    } catch (UnrecognizedOptionException ure) {
      System.err.println(ure.getMessage());
      HelpFormatter hf = new HelpFormatter();
      hf.printHelp("sdslc <options> <input filename>", options);
      System.exit(ExitCodes.BAD_COMMAND_LINE);
    }
    catch (Exception e) {
      System.err.println("Could not parse command line");
      //e.printStackTrace();
      HelpFormatter hf = new HelpFormatter();
      hf.printHelp("sdslc <options> <input filename>", options);
      System.exit(ExitCodes.BAD_COMMAND_LINE);
    }
  }

  public boolean outputDotAST() {
    return outputDotAST;
  }

  public void setOutputDotAST(boolean outputDotAST) {
    this.outputDotAST = outputDotAST;
  }

  public boolean outputTextAST() {
    return outputTextAST;
  }

  public void setOutputTextAST(boolean outputTextAST) {
    this.outputTextAST = outputTextAST;
  }

  public CodeGenerator getCodeGenerator() {
    return codeGenerator;
  }

  public void setCodeGenerator(CodeGenerator codeGenerator) {
    this.codeGenerator = codeGenerator;
  }

  public boolean outputVerbose() {
    return outputVerbose;
  }

  public void setOutputVerbose(boolean outputVerbose) {
    this.outputVerbose = outputVerbose;
  }

  public String getInFileName() {
    return inFileName;
  }

  public void setInFileName(String inFileName) {
    this.inFileName = inFileName;
  }

  public String getOutFileName() {
    return outFileName;
  }

  public void setOutFileName(String outFileName) {
    this.outFileName = outFileName;
  }

  public Level getLogLevel() {
    return logLevel;
  }

  public void setLogLevel(Level logLevel) {
    this.logLevel = logLevel;
  }


  public boolean isExternC() {
    return externC;
  }


  public void setExternC(boolean externC) {
    this.externC = externC;
  }


  public boolean isLegacyGPU() {
    return legacyGPU;
  }


  public void setLegacyGPU(boolean legacyGPU) {
    this.legacyGPU = legacyGPU;
  }

}
