/*
 * OvertileOptions.java: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */


package edu.osu.cse.hpcrl.sdsl.util;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Contains space tile size, time tile size, and thread block size for GPU code 
 * generation.
 * 
 * @author Tom Henretty <henretty@cse.ohio-state.edu>
 *
 */
public class OvertileOptions {
  private int timeTileSize;
  private ArrayList<Integer> spaceTileSizes;
  private ArrayList<Integer> threadBlockSizes;

  /**
   * Sets time tile size to 1, does not specify space tile 
   * or thread block sizes.
   */
  public OvertileOptions() {
    timeTileSize = 1;
    spaceTileSizes = new ArrayList<Integer>();
    threadBlockSizes = new ArrayList<Integer>();
  }
  
  /**
   * Builds GPUCodegen options according to the values specified in the gpu 
   * clause of the sdsl begin pragma.
   *  
   * @param pragma  Complete sdsl begin pragma
   * @throws Exception  If any errors are encountered while parsing the pragma
   */
  public OvertileOptions(String pragma) throws Exception {
    // FIXME: Error handling could be better
    String gpuRegex = ".*gpu.*";
    String spaceTileRegex = ".*(tile:[0-9]+(,[0-9]+)*).*";
    String threadBlockRegex = ".*(block:[0-9]+(,[0-9]+)*).*";
    String timeTileRegex = ".*(time:[0-9]*).*";
    
    //===-----------------------------------------------------------------------
    // Bail if the pragma doesn't have a complete gpu clause
    //===-----------------------------------------------------------------------
    if (!pragma.matches(gpuRegex)) {
      throw new Exception("INVALID SDSL PRAGMA: gpu(...) CLAUSE");
    }
    if (!pragma.matches(spaceTileRegex)) {
      throw new Exception("INVALID SDSL PRAGMA: gpu(tile) CLAUSE");
    }
    if (!pragma.matches(threadBlockRegex)) {
      throw new Exception("INVALID SDSL PRAGMA: gpu(block) CLAUSE");
    }
    if (!pragma.matches(timeTileRegex)) {
      throw new Exception("INVALID SDSL PRAGMA: gpu(time) CLAUSE");
    }

    //===-----------------------------------------------------------------------
    // Get overtile options as Strings
    //===-----------------------------------------------------------------------
    Matcher spaceTileMatcher = Pattern.compile(spaceTileRegex).matcher(pragma);
    spaceTileMatcher.find();
    String[] spaceTiles = spaceTileMatcher.group(1).trim().substring(5).split(",");
    
    
    Matcher threadBlockMatcher = Pattern.compile(threadBlockRegex).matcher(pragma);
    threadBlockMatcher.find();
    String[] threadBlocks = threadBlockMatcher.group(1).trim().substring(6).split(",");
    
    Matcher timeTileMatcher = Pattern.compile(timeTileRegex).matcher(pragma);
    timeTileMatcher.find();
    String timeTile = timeTileMatcher.group(1).trim().substring(5);
    

    //===-----------------------------------------------------------------------
    // Convert gpu codegen options to ints and store
    //===-----------------------------------------------------------------------
    spaceTileSizes = new ArrayList<Integer>();
    for (String size : spaceTiles) {
      spaceTileSizes.add(Integer.parseInt(size));
    }

    threadBlockSizes = new ArrayList<Integer>();
    for (String size : threadBlocks) {
      threadBlockSizes.add(Integer.parseInt(size));
    }
    timeTileSize = Integer.parseInt(timeTile);
  }

  /**
   * @return the timeTileSize
   */
  public final int getTimeTileSize() {
    return timeTileSize;
  }

  /**
   * @param timeTileSize the timeTileSize to set
   */
  public final void setTimeTileSize(int timeTileSize) {
    this.timeTileSize = timeTileSize;
  }

  /**
   * @return the spaceTileSizes
   */
  public final ArrayList<Integer> getSpaceTileSizes() {
    return spaceTileSizes;
  }

  /**
   * @param spaceTileSizes the spaceTileSizes to set
   */
  public final void setSpaceTileSizes(ArrayList<Integer> spaceTileSizes) {
    this.spaceTileSizes = spaceTileSizes;
  }

  /**
   * @return the threadBlockSizes
   */
  public final ArrayList<Integer> getThreadBlockSizes() {
    return threadBlockSizes;
  }

  /**
   * @param threadBlockSizes the threadBlockSizes to set
   */
  public final void setThreadBlockSizes(ArrayList<Integer> threadBlockSizes) {
    this.threadBlockSizes = threadBlockSizes;
  }
}
