/*
 * overtile.cpp: This file is part of the SDSLC project.
 *
 * SDSLC: A compiler for high performance stencil computations
 *
 * Copyright (C) 2011-2013 Ohio State University
 *
 * This program can be redistributed and/or modified under the terms
 * of the license specified in the LICENSE.txt file at the root of the
 * project.
 *
 * Contact: P Sadayappan <saday@cse.ohio-state.edu>
 */

 
/*
 * @file: overtile.cpp
 * @author: Tom Henretty <henretty@cse.ohio-state.edu>
 * @author: Justin Holewinski <justin.holewinski@gmail.com>
 */

#include "LibOvertileWrapper.h"

#include "overtile/Parser/SSPParser.h"
#include "overtile/Core/CudaBackEnd.h"

#include "llvm/ADT/OwningPtr.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/DynamicLibrary.h"
#include "llvm/Support/ErrorHandling.h"
#include "llvm/Support/ManagedStatic.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/Support/PrettyStackTrace.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Regex.h"
#include "llvm/Support/Signals.h"
#include "llvm/Support/system_error.h"
#include "llvm/Support/ToolOutputFile.h"

#include <bitset>
#include <jni.h>

#include <iostream>

using namespace llvm;
using namespace llvm::sys;
using namespace overtile;



/*
 * Class:     edu_osu_cse_hpcrl_sdsl_codegen_LibOvertileWrapper
 * Method:    generateCUDAFunctionDefinition
 * Signature: (Ljava/lang/String;IIIIIIIZ)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_edu_osu_cse_hpcrl_sdsl_backend_LibOvertileWrapper_generateCUDAFunctionDefinition
  (JNIEnv *env, jclass javaClass, jstring sspJstring, jintArray blockSizes, jintArray spaceTileSizes, jint time, jboolean legacy, jboolean verb) {


  // Convert java string to LLVM MemoryBuffer
  jboolean isCopy;
  const char *sspChars = env->GetStringUTFChars(sspJstring,&isCopy);
  StringRef sspSR(sspChars);
  MemoryBuffer* mb = MemoryBuffer::getMemBuffer(sspSR);
  OwningPtr<MemoryBuffer> inSSP(mb);

  //
  // Codegen (adapted from overtile/tools/otsc/otsc.cpp)
  //
  OwningPtr<Grid> G;
  SourceMgr       SM;
  SSPParser       P(inSSP.take(), SM);

  // Parse. If we hit an error return a comment string
  if (error_code f = P.parseBuffer()) {
	// FIXME: Throw an exception
    errs() << "Abort due to errors\n";
    return env->NewStringUTF("// OVERTILE ERROR: SSP parse error\n");
  }

  G.reset(P.getGrid());
  CudaBackEnd BE(G.get());


  // Set up thread block, tile sizes
  jsize dims = env->GetArrayLength(blockSizes);
  jint *block = env->GetIntArrayElements(blockSizes,0);
  for (jint i = 0; i < dims; ++i) {
	//errs() << "Set block size " << i << " = " << block[i] << "\n";
	BE.setBlockSize(i, block[i]);
  }

  dims = env->GetArrayLength(spaceTileSizes);
  jint *space = env->GetIntArrayElements(spaceTileSizes,0);
  for (jint i = 0; i < dims; ++i) {
	//errs() << "Set element size " << i << " = " << space[i] << "\n";
	BE.setElements(i, space[i]);
  }


  BE.setTimeTileSize((int)time);
  if ((bool)legacy) {
    StringRef machine("gt200");
    BE.setMachine(machine);
  }

  // Set verbosity and generate code
  BE.setVerbose((bool)verb);
  BE.run();

  // Codegen to string
  std::string codeOut;
  llvm::raw_string_ostream ssCodeOut(codeOut);
  BE.codegen(ssCodeOut);

  return env->NewStringUTF(ssCodeOut.str().c_str());
}


/*
 * Class:     edu_osu_cse_hpcrl_sdsl_codegen_LibOvertileWrapper
 * Method:    generatePrototype
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_edu_osu_cse_hpcrl_sdsl_codegen_LibOvertileWrapper_generateCFunctionPrototype
  (JNIEnv *env, jclass javaClass, jstring sspJstring, jboolean verb) {
	  // Convert java string to LLVM MemoryBuffer
	  jboolean isCopy;
	  const char *sspChars = env->GetStringUTFChars(sspJstring,&isCopy);
	  StringRef sspSR(sspChars);
	  MemoryBuffer* mb = MemoryBuffer::getMemBuffer(sspSR);
	  OwningPtr<MemoryBuffer> inSSP(mb);

	  //
	  // Codegen (adapted from overtile/tools/otsc/otsc.cpp)
	  //
	  OwningPtr<Grid> G;
	  SourceMgr       SM;
	  SSPParser       P(inSSP.take(), SM);

	  // Parse. If we hit an error return a comment string
	  if (error_code f = P.parseBuffer()) {
		// FIXME: Throw an exception
	    errs() << "Abort due to errors\n";
	    return env->NewStringUTF("// OVERTILE ERROR: SSP parse error\n");
	  }

	  G.reset(P.getGrid());
	  CudaBackEnd BE(G.get());

	  // FIXME: Make target machine command line option
	  StringRef machine("gt200");
	  BE.setMachine(machine);

	  // Generate code
	  BE.setVerbose((bool)verb);
	  BE.run();

	  // Codegen to string
	  std::string codeOut = BE.getCanonicalPrototype();
	  return env->NewStringUTF(codeOut.c_str());
}
